from typing import Optional, Union, Callable
from datetime import datetime
from ..exception.base import GatewayTimeout
from ..wrappers import Request, Response

from ..settings.timeout import TimeoutConfigSettings

_settings = TimeoutConfigSettings().fetch()

class TimeoutMiddleware:
    def __init__(
        self
    ):
        self.timeout_seconds: Optional[Union[int, float]] = _settings.get('seconds') or 10
        self.timeout_response: Optional[Union[str, Callable[..., Response]]] = _settings.get('response') or "Gateway Timeout"

    async def __call__(self, request: Request, response: Callable[..., Response]):
        start_time = datetime.now()

        elapsed_time = (datetime.now() - start_time).total_seconds()
        if self.timeout_seconds and elapsed_time < self.timeout_seconds:
            return self._handle_timeout()

        return response

    def _handle_timeout(self) -> Response:
        if isinstance(self.timeout_response, str):
            return GatewayTimeout(self.timeout_response)
        elif callable(self.timeout_response):
            return self.timeout_response()
        else:
            raise ValueError("Invalid timeout_response format. Provide a string message or a callable.")



from .cors import CORS as CORSMiddleware
from .limiter import RateLimiter as RateLimiterMiddleware
from .static import StaticMiddleware
from .dispatcher import Dispatcher
from .profiler import AquilifyProfiler as Profiler
from .csp import CSPMiddleware
from .proxyfix import ProxyFix as ProxyFix
from .httpsredirect import HTTPSRedirectMiddleware
from .trustedhost import TrustedhostMiddleware
from .gzip import GzipMiddleware
from .logger import LoggingMiddleware
from .timeout import TimeoutMiddleware
from .compression import CompressionMiddleware
from .xfameoption import XFrameOptionsMiddleware
from .csrfmiddleware import CSRFMiddleware as CSRFMiddleware
from .media import MediaMiddleware as MediaMiddleware
from .hstsmiddleware import HSTSMiddleware as HSTSMiddleware
from .admin import ConsoleAPI as TestConsoleAPI
from .conditional_get import ConditionalGetMiddleware as ConditionalGetMiddleware

__all__ = [
    'CORSMiddleware',
    'RateLimiterMiddleware',
    'StaticMiddleware',
    'Dispatcher',
    'Profiler',
    'CSPMiddleware',
    'ProxyFixer',
    'HTTPSRedirectMiddleware',
    'TrustedhostMiddleware',
    'GzipMiddleware',
    'LoggingMiddleware',
    'TimeoutMiddleware',
    'CompressionMiddleware',
    'XFrameOptionsMiddleware',
    'CSRFMiddleware',
    'MediaMiddleware',
    'HSTSMiddleware',
    'TestConsoleAPI',
    'ConditionalGetMiddleware'
]



from __future__ import annotations

import json
import inspect
import warnings
import traceback
import xml.etree.ElementTree as ET

from enum import Enum
from functools import wraps
from inspect import signature
from collections import defaultdict


from typing import (
    Callable,
    Any,
    List,
    Type,
    Dict,
    Optional,
    TypeVar,
    Tuple,
    Awaitable,
    Pattern,
    Union,
    Mapping
)

from aquilify.config import Config
from aquilify.settings import settings
from aquilify.views.urlI8N import urlI8N
from aquilify.responses import HTMLResponse
from aquilify.core.schematic import Schematic
from aquilify.wrappers.reqparser import Reqparser
from aquilify.core.__status import exception_dict
from aquilify.utils.module_loading import import_string
from aquilify.exception.__handler import handle_exception
from aquilify.settings.lifespan import ASGILifespanLoader
from aquilify.exception.base import ImproperlyConfigured

from aquilify.exception.debug import (
    debug_404,
    debug_405
)

from aquilify.exception.base import (
    HTTPException,
    InternalServerError
)

from aquilify.wrappers import (
    Request,
    Response,
    WebSocket,
    WebSocketDisconnect,
    WebSocketState
)

from aquilify.types import (
    ASGIApp,
    Scope,
    Receive,
    Send,
    Lifespan,
    StatefulLifespan
)

from aquilify.core.__globals import (
    Converter,
    routing,
    BaseSettings,
    StageHandler,
    fetchSettingsMiddleware
)



T = TypeVar("T")

_settings = BaseSettings().compator()

class RequestStage(Enum):
    BEFORE: str = 'before'
    AFTER: str = 'after'

class ColoursCode:
    BG_YELLOW = "\033[35m"
    BLUE = "\033[34m"
    GREEN = "\033[32m"
    RED = "\033[31m"
    YELLOW = "\033[33m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    BOLD = "\033[1m"
    RESET = "\033[0m" 

class Aquilify:
    def __init__(
        self
        ) -> None:

        self.routes: List[
            Tuple[
                str,
                List[str],
                Callable[..., Awaitable[T]],
                List[str],
                Pattern,
                Type[T],
                str,
                Dict[str, Any],
            ]
        ] = []
        self.websockets: List[
            Tuple[
                str,
                Callable[..., Awaitable[T]]
            ]
        ] = []
        self._middlewares: List[Callable[..., Awaitable[T]]] = []
        self.middleware_order: List[Tuple[Callable[..., Awaitable[T]], int]] = []
        self.request_transformers: List[Callable[..., Awaitable[Request]]] = []
        self.response_transformers: List[
            Callable[..., Awaitable[Response]]
        ] = []
        self.messages: List[Callable[..., Awaitable[T]]] = []
        self.middleware_groups: Dict[str, List[Callable[..., Awaitable[T]]]] = {}
        self.middleware_activation: Dict[Callable[..., Awaitable[T]], bool] = {}
        self.middleware_dependencies: Dict[Callable[..., Awaitable[T]], List[Callable[..., Awaitable[T]]]] = defaultdict(list)
        self.middleware_exclusions: Dict[Callable[..., Awaitable[T]], List[Callable[..., Awaitable[T]]]] = defaultdict(list)
        self.before_request_handlers: List[Callable[..., Awaitable[T]]] = []
        self.after_request_handlers: List[Callable[..., Awaitable[T]]] = []
        self.startup_handlers: List[Callable[..., Awaitable[Lifespan]]] = []
        self.shutdown_handlers: List[Callable[..., Awaitable[Lifespan]]] = []
        self.grouped_request_stages: Dict[str, Dict[str, List[Callable]]] = {}
        self.error_handlers: Dict[str, Dict[str, List[Callable]]] = {}
        self.excluded_stages: Dict[str, List[Callable]] = {}
        self.config: Callable[..., Awaitable[T, Config, Union[str, dict, bytes, Any]]] = None,
        self.request_stage_handlers: Dict[str, List[Tuple[Callable[..., Awaitable[None]], int, Optional[Callable]]]] = {
            RequestStage.BEFORE.value: [], 
            RequestStage.AFTER.value: []
        }
        self.on_startup: Optional[Union[Callable[..., Awaitable[Any]], List[Callable[..., Awaitable[Any]]]]] = None,
        self.on_shutdown: Optional[Union[Callable[..., Awaitable[Any]], List[Callable[..., Awaitable[Any]]]]] = None,

        self.debug: bool = _settings or False
        self.schematic_id: Optional[str] = None
        self.schematic: Callable[..., Awaitable[T]] = None
        self.exception_handlers: Optional[
            Mapping[
                Any,
                Callable[
                    [Request, Exception],
                    Union[Response, Awaitable[Response]],
                ],
            ]
        ] = None
        self.settings_stage_handler = StageHandler()
        self.settings_stage_handler.process_stage_handlers(self)
        fetchSettingsMiddleware(self)
        self._check_lifespan_settings()
        self._load_exception_handler()

    def errorhandler(self, status_code: int) -> Callable:
        def decorator(handler: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
            self.error_handlers[status_code] = handler
            return handler
        return decorator

    def route(
        self,
        path: str,
        methods: Optional[List[str]] = None,
        response_model: Optional[Type[T]] = None,
        endpoint: Optional[str] = None,
        strict_slashes: bool = True,
    ) -> Callable[..., Awaitable[T]]:
        """
        We no longer document this decorator style API, and its usage is discouraged.
        Instead you should use the following approach:

        >>> ROUTER = [
                rule('/', methods=['GET', 'POST'], endpoint=home)
            ]
        """
        warnings.warn(
            "The `route` decorator is deprecated, and will be removed in version 1.12. ",  # noqa: E501
            DeprecationWarning,
        )
        allowed_methods = ["GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH", "TRACE"]
        try:
            if not path.startswith('/'):
                raise TypeError("Websocket paths must startwith '/'.")
            if methods is not None:
                for method in methods:
                    if method.upper() not in allowed_methods:
                        raise ValueError(f"Invalid HTTP method provided: {method}")

            if methods is None:
                methods = ["GET"]

            def decorator(
                handler: Callable[..., Awaitable[T]]
            ) -> Callable[..., Awaitable[T]]:
                if not (inspect.iscoroutinefunction(handler) or inspect.isasyncgenfunction(handler)):
                    raise TypeError("ASGI can only register asynchronous functions.")
                converted_path, path_regex = Converter()._regex_converter(
                    path, strict_slashes, ''
                )
                self.routes.append(
                    (
                        converted_path,
                        methods,
                        handler,
                        path_regex,
                        response_model,
                        endpoint,
                    )
                )
                return handler

            return decorator
        except Exception as e:
            if self.debug:
                print(e)
            else:
                raise InternalServerError
            
    def _helper_route_setup(self):
        routes_to_add = []

        for route in routing._routes:
            path, methods, handler, strict_slashes, response_model, endpoint = route
            route_tuple = (
                path,
                tuple(methods),
                handler,
                strict_slashes,
                response_model,
                endpoint,
            )

            if route_tuple not in self.routes:
                routes_to_add.append(route_tuple)

        self.routes.extend(routes_to_add)
    
    def add_route(
        self,
        path: str,
        methods: Optional[List[str]] = None,
        response_model: Optional[Type[T]] = None,
        endpoint: Optional[Callable[..., Awaitable[T]]] = None,
        strict_slashes: bool = True
    ) -> None:
        """
        We no longer document this add_route style API, and its usage is discouraged.
        Instead you should use the following approach:

        >>> ROUTING = [
                routing.route('/', methods=['GET', 'POST'], endpoint=home)
            ]
        """
        warnings.warn(
            "The `add_route` is deprecated, and will be removed in version 1.12. ",  # noqa: E501
            DeprecationWarning,
        )
        if endpoint is None:
            raise ValueError("Handler function is required for adding a route.")
        
        if not (inspect.iscoroutinefunction(endpoint) or inspect.isasyncgenfunction(endpoint)):
            raise TypeError("ASGI can only register asynchronous functions.")

        if not path.startswith('/'):
            raise TypeError("Paths must start with '/'.")
        
        if methods is not None:
            allowed_methods = ["GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH", "TRACE"]
            for method in methods:
                if method.upper() not in map(str.upper, allowed_methods):
                    raise ValueError(f"Invalid HTTP method provided: {method}")

        if methods is None:
            methods = ["GET"]

        converted_path, path_regex = Converter()._regex_converter(
            path, strict_slashes, ''
        )
        handler = endpoint
        self.routes.append(
            (
                converted_path,
                methods,
                handler,
                path_regex,
                response_model,
                endpoint,
            )
        )

    def add_websocket_route(
        self,
        path: str,
        handler: Optional[Callable[..., Awaitable[T]]] = None
    ) -> None:
        """
        We no longer document this add_websocket_route style API, and its usage is discouraged.
        Instead you should use the following approach:

        >>> ROUTING = [
                routing.websocket('/ws', endpoint=func)
            ]
        """
        warnings.warn(
            "The `add_websocket_route` is deprecated, and will be removed in version 1.12. ",  # noqa: E501
            DeprecationWarning,
        )
        if handler is None:
            raise ValueError("Handler function is required for adding a websocket route.")
        
        if not (inspect.iscoroutinefunction(handler) or inspect.isasyncgenfunction(handler)):
            raise TypeError("ASGI Websocket can only register asynchronous functions.")
            
        if not path.startswith('/'):
            raise TypeError("Websocket paths must start with '/'.")
            
        compiled_path, path_regex = Converter()._regex_converter(path, False)
        self.websockets.append((compiled_path, path_regex, handler))
    
    def request_stage(
        self,
        stage: Union[Callable[..., Awaitable[T]], str],
        order: int = 0,
        condition: Optional[Callable[..., bool]] = None,
        group: Optional[str] = None,
        exclude: Optional[str] = None,
        inherit: Optional[str] = None
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await func(*args, **kwargs)

            self.request_stage_handlers[stage].append((wrapper, order, condition))
            self.request_stage_handlers[stage] = sorted(self.request_stage_handlers[stage], key=lambda x: x[1])
            
            if group:
                if group not in self.grouped_request_stages:
                    self.grouped_request_stages[group] = {}
                self.grouped_request_stages[group].setdefault(stage, []).append(wrapper)
                
            if exclude:
                self.excluded_stages.setdefault(exclude, set()).add(wrapper)
                
            if inherit:
                self._inherit_from_group(stage, group, inherit)

            return wrapper

        return decorator
    
    def stage_handler(
        self,
        func: Callable,
        stage: Union[Callable[..., Awaitable[Any]], str],
        order: int = 0,
        condition: Optional[Callable[..., bool]] = None,
        group: Optional[str] = None,
        exclude: Optional[str] = None,
        inherit: Optional[str] = None
    ) -> Callable:
        @wraps(func)
        async def wrapper(*args, **kwargs):
            return await func(*args, **kwargs)

        self.request_stage_handlers.setdefault(stage, []).append((wrapper, order, condition))
        self.request_stage_handlers[stage] = sorted(self.request_stage_handlers[stage], key=lambda x: x[1])

        if group:
            if group not in self.grouped_request_stages:
                self.grouped_request_stages[group] = {}
            self.grouped_request_stages[group].setdefault(stage, []).append(wrapper)

        if exclude:
            self.excluded_stages.setdefault(exclude, set()).add(wrapper)

        if inherit:
            self._inherit_from_group(stage, group, inherit)

        return wrapper

    def _inherit_from_group(
        self,
        stage: Union[Callable[..., Awaitable[T]], str],
        group: Optional[str],
        inherit_group: Optional[str]
    ) -> None:
        if group in self.grouped_request_stages and inherit_group in self.grouped_request_stages:
            inherited_stages = self.grouped_request_stages[inherit_group].get(stage, [])
            self.grouped_request_stages[group].setdefault(stage, []).extend(inherited_stages)

    
    async def _execute_request_stage_handlers(
        self,
        stage: Union[Callable[..., Awaitable[T]], str],
        *args: Optional[Union[Callable[..., Awaitable[T]], str]],
        context: Dict[str, Any],
        **kwargs
    ) -> Any:
        handlers = self.request_stage_handlers[stage] + self.grouped_request_stages.get(stage, [])
        if stage in self.excluded_stages:
            handlers = [handler for handler in handlers if handler not in self.excluded_stages[stage]]
        for handler, _, condition in handlers:
            try:
                if not condition or (condition and condition(*args, **kwargs)):
                    func_args = list(signature(handler).parameters.keys())
                    handler_args = args + (context,) if 'context' in func_args else args
                    result = await handler(*handler_args, **kwargs)
                    if result:
                        return result
            except Exception as e:
                if self.debug:
                    await handle_exception(e, *args)
                    print(f"Error in {stage} request stage handler: {e}")
                else:
                    return await self._error_validator(500)
        return None

    def include(
        self,
        schematic: Dict[str, Schematic[ASGIApp]],
        include_middlewares: Optional[bool] = False
    ) -> None:
        for url_prefix, schematic_instance in schematic.items():
            self._process_routes(schematic_instance, url_prefix)
            if include_middlewares:
                self._add_middlewares(schematic_instance)
            self._process_schematic_instance(schematic_instance, url_prefix)

    def _process_routes(self, schematic_instance: Schematic[ASGIApp], url_prefix: str) -> None:
        for route in schematic_instance.routes:
            path, methods, handler, strict_slashes, response_model, endpoint = route

            if path.endswith('/'):
                path = ''  # Simulating the root_path

            converted_path, path_regex = Converter()._regex_converter(url_prefix + path, strict_slashes)

            self.routes.append(
                (
                    converted_path,
                    methods,
                    handler,
                    path_regex,
                    response_model,
                    endpoint,
                )
            )

    def _add_middlewares(self, schematic_instance: Schematic[ASGIApp]) -> None:
        self._middlewares.extend(schematic_instance.middlewares)

    def _process_schematic_instance(self, schematic_instance: Schematic[ASGIApp], url_prefix: str) -> None:
        if schematic_instance.schematic_id is not None:
            self._update_schematic_info(schematic_instance)
            self._print_schematic_info(schematic_instance, url_prefix)

        self._process_websockets(schematic_instance, url_prefix)

    def _update_schematic_info(self, schematic_instance: Schematic[ASGIApp]) -> None:
        self.schematic_id = schematic_instance.schematic_id
        self.schematic = schematic_instance.get_schematic()

    def _print_schematic_info(self, schematic_instance: Schematic[ASGIApp], url_prefix: str) -> None:
        schematic_name = schematic_instance.name
        url_prefix_info = (
            f"at {ColoursCode.BOLD}{ColoursCode.GREEN}{url_prefix}{ColoursCode.RESET} url-prefix"
            if url_prefix
            else "with no specific URL prefix"
        )

        serving_message = (
            f"\n Serving Schematic {ColoursCode.BOLD}{ColoursCode.GREEN}'{schematic_name}'{ColoursCode.RESET} Instance {url_prefix_info}."
        )
        instance_id_message = f"Instance-ID: {ColoursCode.BOLD}{ColoursCode.GREEN}{self.schematic_id}{ColoursCode.RESET}"

        print(serving_message)
        print(instance_id_message)

        paths = [route[0] for route in schematic_instance.routes]
        if paths:
            print(
                f"{ColoursCode.CYAN}Routes{ColoursCode.RESET}: {ColoursCode.GREEN}{paths}{ColoursCode.RESET} \n"
            )
        else:
            print(
                f"{ColoursCode.RED}No HTTP routes found for this schematic.{ColoursCode.RESET} \n"
            )

    def _process_websockets(self, schematic_instance: Schematic[ASGIApp], url_prefix: str) -> None:
        if schematic_instance.websockets:
            print(f"{ColoursCode.BOLD}WebSocket Routes:{ColoursCode.RESET}")

            for path, handler in schematic_instance.websockets:
                full_path = url_prefix + path
                self.websockets.append(
                    (
                        full_path,
                        Converter()._regex_converter(full_path, False)[1],
                        handler
                    )
                )
                print(f"  - Path: {ColoursCode.GREEN}{full_path}{ColoursCode.RESET}")
                print(f"    Handler: {handler.__name__}")
            
            print()
        else:
            print(f"{ColoursCode.RED}No WebSocket routes found for this schematic.{ColoursCode.RESET}\n")

    def middleware(
        self,
        type: str = "http",
        order: Optional[int] = 0,
        conditions: Optional[List[Callable[..., bool]]] = None,
        group: Optional[str] = None,
        active: bool = True,
        excludes: Optional[Callable[..., Awaitable[T]]] = None
    ) -> Callable:
        if type not in ["http"]:
            raise ValueError(f"Invalid middleware type: {type}. Supported types are 'http'.")

        def decorator(middleware_func: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
            middleware_entry = {
                "middleware": middleware_func,
                "type": type,
                "order": order,
                "conditions": conditions,
                "group": group,
                "excludes": excludes,
                "type": type
            }
            self._middlewares.append(middleware_entry)
            self._middlewares.sort(key=lambda m: m["order"])
            if group:
                self.middleware_groups[group].append(middleware_func)
            self.middleware_activation[middleware_func] = active
            if excludes:
                self.middleware_exclusions[excludes].append(middleware_func)
            return middleware_func
        return decorator

    def add_middleware( 
        self,
        middleware: Callable[..., Awaitable[T]],
        order: Optional[int] = 0,
        conditions: Optional[List[Callable[..., bool]]] = None,
        group: Optional[str] = None,
        active: bool = True,
        excludes: Optional[Callable[..., Awaitable[T]]] = None,
        type: str = 'http'
    ) -> None:
        if type not in ["http"]:
            raise ValueError(f"Invalid middleware type: {type}. Supported types are 'http'.")
        middleware_entry = {
            "middleware": middleware,
            "order": order,
            "conditions": conditions,
            "group": group,
            "excludes": excludes,
            "type": type
        }
        self._middlewares.append(middleware_entry)
        self._middlewares.sort(key=lambda m: m["order"])
        if group:
            self.middleware_groups[group].append(middleware)
        self.middleware_activation[middleware] = active
        if excludes:
            self.middleware_exclusions[excludes].append(middleware)

    async def apply_middlewares(
        self, request: Request, response: Response
    ) -> Response:
        executed_middlewares = set()

        for middleware_entry in self._middlewares:
            middleware = middleware_entry["middleware"]
            conditions = middleware_entry.get("conditions")
            group = middleware_entry.get("group")
            excludes = middleware_entry.get("excludes")

            if group and middleware not in self.middleware_groups[group]:
                continue

            if not self.middleware_activation.get(middleware, True):
                continue

            if conditions:
                if not all(cond(request) for cond in conditions):
                    continue

            if excludes:
                exclusions = self.middleware_exclusions[excludes]
                if any(exc in executed_middlewares for exc in exclusions):
                    continue

            response = await middleware(request, response)
            executed_middlewares.add(middleware)

        return response
    
    async def handle_request(
        self,
        scope: Dict[str, Scope],
        receive: Callable[..., Awaitable[Receive]],
        send: Callable[..., Awaitable[Send]],
    ) -> None:
        path = scope.get("path", "/")
        method = scope.get("method", "GET")
        request = Request(scope, receive)
        response = None
        context: Dict[str, List[Callable[..., Awaitable[T]]]] = {}

        try:
            self._helper_route_setup()
            allowed_methods = set()

            if not self.routes:
                if self.debug:
                    response = HTMLResponse(urlI8N())
                else:
                    response = HTMLResponse("<h1>Welcome to Aquilify, Your installation successful.</h1><p>You have debug=False in you Aquilify settings, change it to True in use of development for better experiance.")
            for (
                route_pattern,
                methods,
                handler,
                path_regex,
                response_model,
                endpoint,
            ) in self.routes:
                match = path_regex.match(path)
                if match:
                    if not methods or method.upper() in map(str.upper, methods):
                        path_params = match.groupdict()
                        processed_path_params = {key: self._convert_value(value) for key, value in path_params.items()}
                        request.path_params = processed_path_params

                        await self._execute_request_stage_handlers(RequestStage.BEFORE.value, request, context=context)

                        await self._context_processer(request)
                        scope['context'] = request.context ## context manager ..., Awaitable

                        for transformer in self.request_transformers:
                            request = await transformer(request)

                        handler_params = inspect.signature(handler).parameters

                        if 'request' in handler_params:
                            parser = Reqparser()

                            if 'parser' in handler_params:
                                response = await handler(request, parser=parser, **request.path_params)
                            else:
                                if request.path_params:
                                    valid_path_params = {key: value for key, value in request.path_params.items() if key in handler_params}
                                    response = await handler(request, **valid_path_params)
                                else:
                                    response = await handler(request)
                        else:
                            handler_kwargs = {param.name: request.path_params[param.name] for param in handler_params.values() if param.name in request.path_params}

                            if handler_kwargs and not request.path_params:
                                raise ValueError("Handler kwargs provided without request.path_params")

                            response = await handler(**handler_kwargs)

                        response = await self._process_response(response, handler.__name__)

                        if response_model:
                            response = self._validate_and_serialize_response(
                                response, response_model
                            )

                        break

                    else:
                        allowed_methods.update(methods)

            if response is None:
                if allowed_methods:
                    response = await self._error_validator(405, request, allowed_methods)
                else:
                    response = await self._error_validator(404, request)

            for transformer in self.response_transformers:
                response = await transformer(response)

            response = await self.apply_middlewares(request, response)
            if not isinstance(response, (Response, Awaitable)):
                raise ValueError("Middleware must return a Response object or Awaitable[Response]")
            
            await self._execute_request_stage_handlers(RequestStage.AFTER.value, request, response, context=context)

        except Exception as e:
            response = await self._process_exception(e, request)

        await response(scope, receive, send)

    async def _context_processer(self, request: Request):
        request.context['_request'] = request
        request.context['_app'] = self

    async def _process_exception(self, e, request) -> Response:
        reversed_exception_dict = {v: k for k, v in exception_dict.items()}
        if type(e) in reversed_exception_dict:
            status_code = reversed_exception_dict[type(e)]
            response = await self._error_validator(status_code, request)
            return response
        else:
            if self.debug:
                response = await handle_exception(e, request)
            elif self.exception_handlers:
                response = await self.exception_handlers(e, request)
            else:
                response = await self._error_validator(500)
        return response
    
    def _convert_value(self, value):
        if isinstance(value, int):
            return int(value)
        elif isinstance(value, str):
            if value.isdigit():
                return int(value)
            try:
                return float(value)
            except ValueError:
                return value
        return str(value)

    async def _process_response(self, response, caller_function) -> Response:
        caller_function_name = caller_function
        if isinstance(response, str):
            if response.startswith("<"):
                try:
                    ET.fromstring(response)
                    response = Response(response, content_type='application/xml')
                except ET.ParseError:
                    response = Response(response, content_type='text/html')
            else:
                response = Response(response, content_type='text/plain')
        elif isinstance(response, dict):
            response = Response(content=json.dumps(response), content_type='application/json')
        elif isinstance(response, tuple) and len(response) == 2 and isinstance(response[1], int):
            if isinstance(response[0], str):
                if response[0].startswith("<"):
                    try:
                        ET.fromstring(response[0])
                        response = Response(response[0], content_type='application/xml', status_code=response[1])
                    except ET.ParseError:
                        response = Response(response[0], content_type='text/html', status_code=response[1])
                else:
                    response = Response(response[0], content_type='text/plain', status_code=response[1])
            elif isinstance(response[0], dict):
                response = Response(content=json.dumps(response[0]), content_type='application/json', status_code=response[1])
            elif isinstance(response[0], list):
                def handle_nested(item):
                    return item if isinstance(item, (str, bytes)) else json.dumps(item)

                processed_list = [handle_nested(item) for item in response[0]]
                
                if all(isinstance(item, (str, bytes)) for item in processed_list):
                    response = Response(content=json.dumps(processed_list), content_type='application/json', status_code=response[1])
                else:
                    response = InternalServerError("Unable to process the list structure")
        elif isinstance(response, list):
            def handle_nested(item):
                return item if isinstance(item, (str, bytes)) else json.dumps(item)

            processed_list = [handle_nested(item) for item in response]
            
            if all(isinstance(item, (str, bytes)) for item in processed_list):
                response = Response(content=json.dumps(processed_list), content_type='application/json')
            else:
                response = InternalServerError("Unable to process the list structure")
        elif not isinstance(response, Response):
            received_type = type(response).__name__
            expected_types = ", ".join([typ.__name__ for typ in [str, dict, Response]])
            raise TypeError(f"Function '{caller_function_name}': Invalid response type: Received {received_type}. Expected types are {expected_types}.")
        return response

    async def _error_validator(self, error_code, *args):
        if error_code in self.error_handlers:
            if error_code == 500:
                error_handler = self.error_handlers[error_code]
                response = await error_handler() if not args else await error_handler(*args)
            else:
                error_handler = self.error_handlers[error_code]
                response = await error_handler(*args) if args else await error_handler()

            if isinstance(response, str):
                return Response(content=response, content_type="text/plain", status_code=error_code)
            elif isinstance(response, dict):
                return Response(content=json.dumps(response), status_code=error_code, content_type='application/json')
            elif isinstance(response, Response):
                return response
            else:
                received_type = type(response).__name__
                expected_types = ", ".join([typ.__name__ for typ in [str, dict, Response]])
                raise HTTPException(f"Invalid response type: Received {received_type}. Expected types are {expected_types}")
        
        if error_code in exception_dict:
            if error_code == 404 and self.debug:
                request: Request = args[0]
                return Response(debug_404(routing._links, { 'path': request.path, 'client_ip': request.remote_addr, 'user_agent': request.user_agent, 'url': request.url, 'method': request.method }), 404, content_type='text/html')
            elif error_code == 405 and self.debug:
                request: Request = args[0]
                return Response(debug_405({ 'method': request.method, "url": request.url, "client_ip": request.remote_addr, "user_agent": request.user_agent, "allowed_method": args[1]}), 405, content_type='text/html')
            else:
                return exception_dict[error_code]()
        else:
            raise TypeError('Unsupported error type! : {}'.format(error_code))

    def _validate_and_serialize_response(
        self, response: Response, response_model: Type[T]
    ) -> Response:
        if not isinstance(response, response_model):
            raise ValueError(
                f"Response does not match the expected model {response_model.__name__}"
            )
        return Response(content=response.dict(), content_type="application/json")
    
    def websocket(
        self,
        path: str
    ) -> Callable[..., Awaitable[T]]:
        """
        We no longer document this decorator style API, and its usage is discouraged.
        Instead you should use the following approach:

        >>> ROUTING = [
                routing.websocket('/ws', endpoint=func)
            ]
        """
        warnings.warn(
            "The `decorator` is deprecated, and will be removed in version 1.12. ",  # noqa: E501
            DeprecationWarning,
        )
        def decorator(
            handler: Callable[..., Awaitable[T]]
        ) -> Callable[..., Awaitable[T]]:
            if not (inspect.iscoroutinefunction(handler) or inspect.isasyncgenfunction(handler)):
                raise TypeError("ASGI Websocket, can only register asynchronous functions.")
            if not path.startswith('/'):
                raise TypeError("Websocket paths must startwith '/'.")
            compiled_path, path_regex = Converter()._regex_converter(path, False)
            self.websockets.append((compiled_path, path_regex, handler))
            return handler

        return decorator

    async def _websocket_handler(
        self,
        scope: Dict[str, Scope],
        receive: Callable[..., Awaitable[Receive]],
        send: Callable[..., Awaitable[Send]],
    ) -> None:
        ws = WebSocket(scope, receive, send)
        try:
            await self._websocket_routes(ws)
            await self._helper_websocket_routes(ws)
        except WebSocketDisconnect as e:
            await ws.close(code=e.code, reason=e.reason)
        except RuntimeError as e:
            if ws.client_state != WebSocketState.CONNECTED:
                await ws.close(code=403, reason="Connection rejected")
            else:
                await ws.close(code=1011, reason="Unexpected condition")
        except Exception as e:
            await ws.send_text(f"Error: {str(e)}")
            if ws.application_state != WebSocketState.CONNECTED:
                await ws.close(code=1006, reason="Connection closed unexpectedly")
            else:
                await ws.close(code=1011, reason="Unexpected condition")

    async def _websocket_routes(self, ws: WebSocket) -> None:
        for path, path_regex, handler in self.websockets:
            match = path_regex.match(ws.path)
            if match:
                ws.path_params = match.groupdict()
                response = await handler(ws, **ws.path_params)
                if not isinstance(response, WebSocket):
                    received_type = type(response).__name__
                    expected_types = ", ".join([typ.__name__ for typ in [WebSocket]])
                    raise TypeError(f"Invalid response type: Received {received_type}. Expected types are {expected_types}.")
                return response
            
    async def _helper_websocket_routes(self, ws: WebSocket) -> None:
        for path, path_regex, handler in routing._websockets:
            match = path_regex.match(ws.path)
            if match:
                ws.path_params = match.groupdict()
                response = await handler(ws, **ws.path_params)
                if not isinstance(response, WebSocket):
                    received_type = type(response).__name__
                    expected_types = ", ".join([typ.__name__ for typ in [WebSocket]])
                    raise TypeError(f"Invalid response type: Received {received_type}. Expected types are {expected_types}.")
                return response
            
    def _load_exception_handler(self) -> Optional[
            Mapping[
                Any,
                Callable[
                    [Request, Exception],
                    Union[Response, Awaitable[Response]],
                ],
            ]
        ]:
        try:
            exception_handler: Optional[str] = getattr(settings, 'EXCEPTION_HANDLER', None)
            if exception_handler is not None:
                if not isinstance(exception_handler, str):
                    raise ImproperlyConfigured("Invalid exception_handler type, excepted string")
            if exception_handler:
                _handler = import_string(exception_handler)
                if inspect.isclass(_handler):
                    self.exception_handlers = _handler()
                else:
                    if not (inspect.iscoroutinefunction(_handler) or inspect.isasyncgenfunction(_handler)):
                        raise TypeError("ASGI can only register asynchronous functions or class of Exception handler.")
                    self.exception_handlers = _handler
            else:
                self.exception_handlers = None 
        except ImproperlyConfigured as config_error:
            raise ImproperlyConfigured(f"Error loading {exception_handler} Exception handler: {config_error}")
        except ImportError as import_error:
            raise ImproperlyConfigured(f"Error importing {exception_handler} Exception handler: {import_error}")
        except Exception as e:
            raise ImproperlyConfigured(f"An unexpected error occurred: {e}")
        
    async def __call__(
        self, scope: Dict[str, Scope], receive: Callable[..., Awaitable[Receive]], send: Callable[..., Awaitable[Send]]
    ) -> None:
        scope['app'] = self
        if self.schematic is not None:
            scope['schematic'] = self.schematic
        if scope['type'] == 'http':
            await self._http(scope, receive, send)
        elif scope['type'] == 'websocket':
            await self._websocket_handler(scope, receive, send)
        elif scope['type'] == 'lifespan':
            await self._lifespan(scope, receive, send)

    async def _http(
        self,
        scope: Dict[str, Scope],
        receive: Callable[..., Awaitable[Receive]],
        send: Callable[..., Awaitable[Send]]
    ) -> None:
        try:
            await self.handle_request(scope, receive, send)
        except Exception as e:
            if self.debug:
                request = Request(scope, receive, send)
                response = await handle_exception(e, request)
            elif self.exception_handlers:
                request = Request(scope, receive, send)
                response = await self.exception_handlers(e, request)
            else:
                response = await self._error_validator(500)
            await response(scope, receive, send)

    async def _lifespan(
        self,
        scope: Dict[str, Scope],
        receive: Callable[..., Awaitable[Receive]],
        send: Send
    ) -> Lifespan:
        
        started = False
        event: StatefulLifespan = await receive() 
        try:
            if event['type'] == 'lifespan.startup':
                await self._startup_handlers()
                await send({"type": "lifespan.startup.complete"})
                started = True
                await receive()
        except BaseException:
            exc_text = traceback.format_exc()
            if started:
                await send({"type": "lifespan.shutdown.failed", "message": exc_text})
            else:
                await send({"type": "lifespan.startup.failed", "message": exc_text})
            raise
        else:
            await self._shutdown_handlers()
            await send({"type": "lifespan.shutdown.complete"})

    async def _startup_handlers(self) -> None:
        for handler in self.startup_handlers:
            if not (inspect.iscoroutinefunction(handler) or inspect.isasyncgenfunction(handler)):
                raise TypeError("ASGI can only register asynchronous lifespan functions.")
            await handler()

    async def _shutdown_handlers(self) -> None:
        for handler in self.shutdown_handlers:
            if not (inspect.iscoroutinefunction(handler) or inspect.isasyncgenfunction(handler)):
                raise TypeError("ASGI can only register asynchronous lifespan functions.")
            await handler()

    def event(self, event_type: str) -> Callable:
        def decorator(handler: Callable[..., Awaitable[Any]]) -> Callable[..., Awaitable[Any]]:
            if event_type == 'startup':
                self.startup_handlers.append(handler)
            elif event_type == 'shutdown':
                self.shutdown_handlers.append(handler)
            return handler
        return decorator

    def _check_events(self, on_startup: Lifespan, on_shutdown: Lifespan):
        if on_startup:
            if isinstance(on_startup, list):
                self.startup_handlers.extend(on_startup)
            else:
                self.startup_handlers.append(on_startup)

        if on_shutdown:
            if isinstance(on_shutdown, list):
                self.shutdown_handlers.extend(on_shutdown)
            else:
                self.shutdown_handlers.append(on_shutdown)
                
    def _check_lifespan_settings(self):
        _settings = ASGILifespanLoader().load_asgi_lifespans()
        for lifespan in _settings:
            if lifespan.get('event') == 'startup':
                self.startup_handlers.append(lifespan.get('origin'))
            elif lifespan.get('event') == 'shutdown':
                self.shutdown_handlers.append(lifespan.get('origin'))
            else:
                raise HTTPException('Invalid event type! {} use either startup or shutdown.'.format(lifespan.get('origin')))



import email
from io import BytesIO

class FormParser:
    def __init__(self, request):
        self.request = request

    async def parse(self):
        content_type = await self.request.header('content-type', '')

        if 'multipart/form-data' in content_type:
            boundary = content_type.split('boundary=')[-1]
            form_data = await self.parse_multipart_form_data(boundary)
        else:
            form_data = await self.parse_url_encoded_form_data()

        return form_data

    async def parse_multipart_form_data(self, boundary):
        form_data = {}
        body = await self.request.body()
        message = email.message_from_bytes(body, _class=email.parser.BytesParser)
        for part in message.walk():
            if part.is_multipart():
                continue
            content_disposition = part.get("Content-Disposition")
            if content_disposition:
                _, params = email.header.decode_header(content_disposition)
                for param, value in params:
                    if param.lower() == 'name':
                        field_name = value
                        if field_name not in form_data:
                            form_data[field_name] = []
                    if param.lower() == 'filename':
                        filename = value
                        field_data = {'filename': filename, 'content': part.get_payload(decode=True)}
                        form_data[field_name].append(field_data)
            else:
                continue

        return form_data

    async def parse_url_encoded_form_data(self):
        body = await self.request.body()
        form_data = {}
        data = body.decode('utf-8')
        form_fields = data.split('&')
        for field in form_fields:
            key, value = field.split('=')
            key = key.replace("+", " ")
            value = value.replace("+", " ")
            form_data[key] = value
        return form_data



import os
import re
import shutil
import argparse
import time
import subprocess
import configparser

from .fax import settings, views, asgi, __root__, routing, ax, tools, models
from .fax.app import main as mainview

from .insecure import ax_insecure_key

class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class AxStartup:
    VERSION = ax.AX_VERSION
    CONFIG_FILE = 'config.cfg'

    def __init__(self):
        self.parser = self._setup_arg_parser()
        
    def measure_time(self, func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            func(*args, **kwargs)
            end_time = time.time()
            elapsed_time = end_time - start_time
            return elapsed_time
        return wrapper
    
    def create_file(self, file_path, content=None):
        try:
            with open(file_path, 'a') as f:
                if content:
                    f.write(content)
        except Exception as e:
            raise e

    def _setup_arg_parser(self):
        parser = argparse.ArgumentParser(description="Create an app with specified name and copy project files.")
        parser.add_argument("-v", "--version", action="store_true", help="Display Aquilify version")
        subparsers = parser.add_subparsers(dest="command", help="Available commands")

        create_app_parser = subparsers.add_parser("create-app", help="Create an app")
        create_app_parser.add_argument("app_name", nargs='?', default=None, help="Name of the app to be created (optional)")

        run_server_parser = subparsers.add_parser("runserver", help="Run the server")

        return parser

    def _print_colored(self, message, color):
        print(f"{color}{message}{Colors.ENDC}")

    def _create_app(self, app_name):
        app_path = os.path.join(os.getcwd(), app_name)
        try:
            if os.path.exists(app_path):
                self._handle_existing_app(app_path, app_name)
            else:
                current_dir = os.getcwd()
                project_dir = os.path.join(current_dir, app_name)
                os.makedirs(project_dir, exist_ok=True)
                self._copy_project_files(app_name, app_path)
        except Exception as e:
            self._print_colored(f"Error: {e}", Colors.FAIL)

    def _handle_existing_app(self, app_path, app_name):
        overwrite = str(input(f"{Colors.WARNING}App folder '{app_name}' already exists. Do you want to overwrite it?{Colors.ENDC} "))
        
        if overwrite in ("YES", 'Yes', 'Y', 'y'):
            self._overwrite_existing_app(app_path, app_name)
        else:
            self._print_bold("Operation aborted. Please choose a different app name.")

    def _overwrite_existing_app(self, app_path, app_name):
        try:
            shutil.rmtree(app_path)
            current_dir = os.getcwd()
            project_dir = os.path.join(current_dir, app_name)
            os.makedirs(project_dir, exist_ok=True)
            self._copy_project_files(app_name, app_path)
            self._print_colored(f"Project '{app_name}' overwritten successfully.", Colors.OKGREEN)
        except (OSError, shutil.Error) as e:
            self._print_colored(f"Failed to overwrite app '{app_name}': {e}", Colors.FAIL)

    def _copy_project_files(self, app_name, app_path):
        try:
            self._copy_package_files(app_name, app_path)
                        
            self._create_config_file(app_path)
            self._create_packlib_file(app_path)
            
            config_file = os.path.join(app_path, 'config.cfg')

            if os.path.exists(config_file):
                self._update_config_file(config_file, app_name, app_path)
            print(f"\nProject '{Colors.OKGREEN}{app_name}{Colors.ENDC}' created successfully. \n")
        except (OSError, shutil.Error) as e:
            self._print_colored(f"Failed to create app '{app_name}': {e}", Colors.FAIL)

    def _create_config_file(self, app_path):
        config_data = """[Aquilify]
PATH = ""
PROJECT_NAME = ""
VERSION = ""
COMPILER_PATH = ""
SETTINGS = ""

[ASGI_SERVER]
SERVER = NETIX
HOST = 127.0.0.1
PORT = 8000
DEBUG = True
RELOAD = False
INSTANCE = asgi:application
"""
        config_file_path = os.path.join(app_path, 'config.cfg')
        self._write_to_file(config_file_path, config_data, "Config file")

    def _create_packlib_file(self, app_path):
        packlib_data = """# This is an aquilify environment which is used to connect the application with secret variables.
# This is only built for system usage so, don't make any changes to the default configuration.

# Warning - If you make any changes to the default environment, it may affect your application and even break the routing.

# Inbuilt environment module of LxEnvironment...

import { environment } from 'LxEnvironment.env'

# Feed the export function to builder and configure the secret environment variables.

environment.export => (builder) = {
    "sysMenSecretKey": "str(base64.encode('utf-8'))",
    "sysEnvironmentPath": "os.path.join('/{folder}/{project}', '.aquilify')",
    "sysEnvironmentSettings": "settings.py",
    "__version__": "float(1.14)",
    "__controller__": "aquilify.core.application",
    "__name__": "aquilify"
}
"""
        packlib_file_path = os.path.join(app_path, 'packlib.lxe')
        self._write_to_file(packlib_file_path, packlib_data, "Packlib file")

    def _write_to_file(self, file_path, data, file_type):
        try:
            with open(file_path, 'w') as file:
                file.write(data)
            print(f"{file_type} {file_path} ...{Colors.OKGREEN}200{Colors.ENDC}")
        except Exception as e:
            self._print_colored(f"Failed to create {file_type.lower()}: {e}", Colors.FAIL)
    
    def _copy_package_files(self, app_name, app_path):
        try:
            files = ['__init__.py', 'views.py', 'settings.py', 'asgi.py', 'models.py', 'tools.py', 'routing.py', '__root__.py', 'config.cfg', 'packlib.lxe']
            total_time = 0

            for file in files:
                file_path = os.path.join(app_path, file)
                try:
                    content = None
                    if file == 'settings.py':
                        content = settings.SETTINGS % ax_insecure_key()
                    elif file == 'views.py':
                        content = views.VIEWS
                    elif file == '__root__.py':
                        content = __root__.ROOT
                    elif file == 'routing.py':
                        content = routing.ROUTING
                    elif file == 'asgi.py':
                        content = asgi.ASGI
                    elif file == 'tools.py':
                        content = tools.TOOLS
                    elif file == 'models.py':
                        content = models.MODLES

                    self.create_file(file_path, content)
                    elapsed_time = self.measure_time(self.create_file)(file_path)
                    total_time += elapsed_time

                    print(f"Created file: {file_path} ...{Colors.OKGREEN}200{Colors.ENDC}")
                except Exception as e:
                    raise e
            for folder_name in ['app', 'lifespan']:
                folder_path = os.path.join(app_path, folder_name)
                os.makedirs(folder_path, exist_ok=True)
                print(f"Created folder: {folder_path} ...{Colors.OKGREEN}200{Colors.ENDC}")
                
                if folder_name == 'app':
                    main_py_path = os.path.join(folder_path, 'main.py')
                    self.create_file(main_py_path, mainview.SCHEMATIC)
                    self.create_file(os.path.join(folder_path, '__init__.py'))
                    print(f"Created folder: {main_py_path} ...{Colors.OKGREEN}200{Colors.ENDC}")
                elif folder_name == 'lifespan':
                    self.create_file(os.path.join(folder_path, '__init__.py'))
                    print(f"Created file: {os.path.join(folder_path, '__init__.py')} ...{Colors.OKGREEN}200{Colors.ENDC}")
                    
        except Exception as e:
            print(f"An error occurred: {e}")

    def _update_config_file(self, config_file, app_name, project_path):
        try:
            config = configparser.ConfigParser()
            config.read(config_file)

            if 'Aquilify' in config:
                config.set('Aquilify', 'PATH', project_path)
                config.set('Aquilify', 'VERSION', '1.14')
                config.set('Aquilify', 'COMPILER_PATH', project_path + '/.aquilify')
                config.set('Aquilify', 'PROJECT_NAME', app_name)
                config.set('Aquilify', 'SETTINGS', project_path  + '/settings.py')

                with open(config_file, 'w') as configfile:
                    config.write(configfile)
                
                print(f"Update {config_file} ...{Colors.OKGREEN}200{Colors.ENDC}")
            else:
                self._print_colored("Aquilify section not found in the config file.", Colors.WARNING)
        except Exception as e:
            self._print_colored(f"Failed to update config file: {e}", Colors.FAIL)
            
    def _install_electrus(self):
        try:
            subprocess.run(["pip", "install", "electrus"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            print(f'Installing https://pypi.org/project/electrus/ ...{Colors.OKGREEN}200{Colors.ENDC}')
            print(f"Electrus installed successfully ...{Colors.OKGREEN}200{Colors.ENDC}")
        except subprocess.CalledProcessError as e:
            self._print_colored(f"Failed to install Electrus: {e.stderr.decode('utf-8')}", Colors.FAIL)

    def _validate_app_name(self, answers, current_app_name):
        app_path = os.path.join(os.getcwd(), current_app_name)
        if os.path.exists(app_path):
            return f"An app with the name '{current_app_name}' already exists. Please choose a different name."
        return True

    def start(self):
        args = self.parser.parse_args()

        if args.version:
            for version in self.VERSION:              
                print(version)
            return  

        if args.command == "create-app":
            self._print_colored(ax.AQUILIFY_THEME, Colors.OKGREEN)
            if args.app_name:
                self._create_app(args.app_name)
            else:
                self._create_app_basic()
        elif args.command == "runserver":
            self._run_server()
        else:
            self.parser.print_help()
        
        return

    def _run_server(self):
        try:
            config = configparser.ConfigParser()
            config.read(self.CONFIG_FILE)

            if 'ASGI_SERVER' in config:
                netix_config = config['ASGI_SERVER']
                server = netix_config.get('SERVER', 'NETIX')
                host = netix_config.get('HOST', '127.0.0.1')
                port = netix_config.get('PORT', '8000')
                debug = netix_config.getboolean('DEBUG', fallback=False)
                reload = netix_config.getboolean('RELOAD', fallback=False)
                instance = netix_config.get('INSTANCE', 'routing:app')

                if server == 'NETIX':
                    cmd = ['netix']
                elif server == 'UVICORN':
                    cmd = ['uvicorn']
                else:
                    self._print_colored("Unsupported ASGI Server {} by Aquilify. Visit http://aquilify.vvfin.in/project/config/server for configuration details.".format(server), Colors.FAIL)
                    cmd = None

                if cmd:
                    if server == 'NETIX' and debug:
                        cmd.append('--debug')

                    if reload:
                        cmd.append('--reload')

                    if port:
                        cmd.append('--port')
                        cmd.append(port)

                    if host:
                        cmd.append('--host')
                        cmd.append(host)

                    cmd.append(instance.replace('__instance__', 'routing:app'))
                    subprocess.run(cmd)
            else:
                self._print_colored("Oops! Either you are not in the project directory or ASGI_SERVER block may not be configured properly. Visit http://aquilify.vvfin.in/project/config/errors for help.", Colors.WARNING)

        except Exception as e:
            self._print_colored(f"An error occurred while running the server, either {server} isn't installed or sys error!: {e}", Colors.FAIL)
        except KeyboardInterrupt as e:
            pass

    def _print_bold(self, message):
        print(f"{Colors.BOLD}{message}{Colors.ENDC}")
        
    def _create_app_basic(self):
        app_name = input("Enter the name of the app: ").strip()
        if not app_name:
            self._print_colored("App name can't be empty!", Colors.FAIL)
            return self._create_app_basic()

        if not re.match(r'^[\w\-.]+$', app_name):
            self._print_colored("Invalid project name. Use only letters, numbers, underscores, hyphens, and periods.", Colors.FAIL)
            return

        if not self._validate_app_name('', app_name):
            return

        setup_db = input("Would you like to set up Electrus database (Yes/No): ").strip().lower()
        
        if setup_db not in ("yes", "no"):
            self._print_colored("Invalid value received: {}. Use either Yes or No.".format(setup_db), Colors.FAIL)
            return

        if setup_db == "yes":
            self._print_colored("Installing Electrus... \n", Colors.OKGREEN)
            try:
                self._install_electrus()
            except Exception as e:
                self._print_colored(f"Error while installing Electrus: {e}", Colors.FAIL)

        self._create_app(app_name)
            
def main():
    cls = AxStartup().start()
    return cls

if __name__ == "__main__":
    main()


from typing import Any, Awaitable, Callable
from aquilify.wrappers import Request, Response

class CSRFConfigurationError(Exception):
    pass

class CSRFMiddleware:
    def __init__(self) -> None:
        pass

    async def __call__(
        self,
        request: Request,
        response: Response
    ) -> Response:
        csrf_token_key = '_csrf_token'
        csrf_view_key = '_csrf_view'

        if request.method in ['GET']:
            if csrf_token_key not in request.context:
                return response
            
            if csrf_view_key not in request.context:
                return response
            
            csrf_token = request.context.get(csrf_token_key)
            csrf_view = request.context.get(csrf_view_key)

            try:
                await self._inject_csrf_token(csrf_view, response, csrf_token)
                response.headers['X-CSRF-TOKEN'] = csrf_token
                return response
            except CSRFConfigurationError as e:
                raise e
        return response

    async def _inject_csrf_token(
        self,
        csrf_view: Callable[..., Awaitable[None]],
        response: Response,
        csrf_token: Any
    ) -> None:
        await csrf_view.inject_csrf_token(response, csrf_token)



import importlib.util

class RateLimitConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_limit_config(self):
        try:
            settings = self._load_settings_module()
            limit = getattr(settings, 'RATE_LIMITER_LIMIT', 60)
            window = getattr(settings, 'RATE_LIMITER_WINDOW', 60)
            exempt_paths = getattr(settings, 'RATE_LIMITER_EXEMPT_PATHS', [])
            headers = getattr(settings, 'RATE_LIMITER_HEADERS', {})
            dynamic_limit = getattr(settings, 'RATE_LIMITER_DYNAMIC_LIMIT', {})
            concurrency_control = getattr(settings, 'RATE_LIMITER_CONCURRENCY_CONTROL', False)
            expiry = getattr(settings, 'RATE_LIMITER_EXPIRY', {})
            monitoring_callback = self._fetch_response_handler_functions("RATE_LIMITER_MONITORING_CALLBACK") or None
            whitelist = getattr(settings, 'RATE_LIMITER_WHITELIST', [])
            blacklist = getattr(settings, 'RATE_LIMITER_BLACKLIST', [])
            endpoint = getattr(settings, 'RATE_LIMITER_BY_ENDPOINT', {})
            adjustments = getattr(settings, 'RATE_LIMITER_ADJUSTMENTS', {})
            expiry_policies = getattr(settings, 'RATE_LIMITER_EXPIRY_POLICIES', {})
            ip_rate_limits = getattr(settings, 'RATE_LIMITER_IP_RATE_LIMITS', {})
            bypass_token = getattr(settings, 'RATE_LIMITER_BYPASS_TOKEN', {})
            granular = getattr(settings, 'RATE_LIMITER_GRANULAR', {})
            automated_asjustments = self._fetch_response_handler_functions("RATE_LIMITER_AUTOMATED_ADJUSTMENTS") or None
            advance_monitiring_and_reporting = self._fetch_response_handler_functions("RATE_LIMITER_ADVANCE_MONITORING_AND_REPORTING") or None
            granular_callback = self._fetch_response_handler_functions("RATE_LIMITER_GRANULAR_CALLBACK") or None

            rate_limit_dict = {
                'limit': limit,
                'window': window,
                'exempt_paths': exempt_paths,
                'headers': headers,
                'dynamic_limit': dynamic_limit,
                'concurrency_control': concurrency_control,
                'expiry': expiry,
                'monitoring_callback': monitoring_callback,
                'whitelist': whitelist,
                'blacklist': blacklist,
                'endpoint': endpoint,
                'adjustments': adjustments,
                'expiry_policies': expiry_policies,
                'ip_rate_limits': ip_rate_limits,
                'bypass_token': bypass_token,
                'granular': granular,
                'automated_adjustments': automated_asjustments,
                'advance_monitoring_and_reporting': advance_monitiring_and_reporting,
                'granular_callback': granular_callback
            }
            return rate_limit_dict
                    
        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings
    
    def _fetch_response_handler_functions(self, object):
        settings = self._load_settings_module()
        
        if settings:
            path_to_callback = getattr(settings, object, None)
            
            if path_to_callback:
                try:
                    parts = path_to_callback.split('.')
                    module_path = '.'.join(parts[:-1])
                    callback_name = parts[-1]
                    callback_module = importlib.import_module(module_path)
                    callback_obj = getattr(callback_module, callback_name)
                    if isinstance(callback_obj, type):
                        callback_obj = callback_obj()
                    
                    return callback_obj
                except (AttributeError, ImportError, ValueError) as e:
                    print(f"Error fetching callback: {e}")
        
        return None

    def fetch(self):
        data = self._fetch_limit_config()
        return data


import re

from typing import Pattern, Tuple

class Converter:
    def _regex_converter(
        self,
        path: str,
        strict_slashes: bool,
        prefix: str = ""
    ) -> Tuple[str, Pattern]:
        
        if strict_slashes:
            pattern = re.sub(r"{([^}]+)}", rf"{prefix}(?P<\1>[^/]+)", path)
        else:
            pattern = re.sub(r"{([^}]+)}", rf"{prefix}(?P<\1>[^/]+/?)", path)
        pattern = f"^{pattern}$"
        path_regex = re.compile(pattern)
        return pattern, path_regex


from typing import Awaitable, List, Optional
from ..wrappers import Request, Response
from ..exception.base import Forbidden
from ..responses import RedirectResponse

from ..settings.trustedhost import TrustedHostConfigSettings

_settings = TrustedHostConfigSettings().fetch()

class TrustedhostMiddleware:
    def __init__(
        self
    ) -> None:
        self.allowed_hosts: List[str] = _settings.get('allowed_hosts') or ["*"]
        self.allow_subdomains: bool = _settings.get('allow_subdomains') or True
        self.redirect_on_fail: bool = _settings.get("redirect_on_fail") or False
        self.redirect_url: Optional[str] = _settings.get('redirect_url') or None
        self.enforce_https: bool = _settings.get('enforce_https') or False
        self.www_redirect: bool = _settings.get('www_redirect') or False

    async def __call__(
        self, request: Request, response: Response
    ) -> Awaitable[Response]:
        try:
            host = request.host
            if not host:
                return await self._handle_fail_response()
            
            if self.www_redirect and host.startswith("www."):
                return self._redirect_to_non_www(request, response)
                
            if self._is_host_allowed(host):
                if self.enforce_https and not request.scheme.startswith("https"):
                    return await self._redirect_to_https(request, response)
            else:
                return await self._handle_fail_response()
                
            return response

        except Exception as e:
            raise

    def _is_host_allowed(self, host: str) -> bool:
        if "*" in self.allowed_hosts:
            return True
        elif host in self.allowed_hosts:
            return True
        elif self.allow_subdomains:
            for allowed_host in self.allowed_hosts:
                if host.endswith("." + allowed_host):
                    return True
        return False

    async def _redirect_to_https(
        self, request: Request, response: Response
    ) -> Response:
        redirect_url = f"https://{request.headers['host']}{request.url.path}"
        if request.url.query:
            redirect_url += f"?{request.url.query}"
        response = RedirectResponse(redirect_url, 301)
        return response
    
    async def _redirect_to_non_www(
        self, request: Request, response: Response
    ) -> Response:
        redirect_url = f"{request.scheme}://{request.headers['host'].replace('www.', '')}{request.url.path}"
        if request.url.query:
            redirect_url += f"?{request.url.query}"
        response = RedirectResponse(redirect_url, 301)
        return response

    async def _handle_fail_response(self) -> Response:
        response = Forbidden("Host not allowed")
        if not response:
            if self.redirect_on_fail and self.redirect_url:
                response = RedirectResponse(self.redirect_url, 307)
            return response
        return response



import os
import re
import json

class LxEnviroment:
    def __init__(self, filename):
        self.filename = self.validate_filename(filename)
        self.filepath = f"{self.filename}.lxe"

    def validate_filename(self, filename):
        # Remove extension if provided
        filename_without_extension = os.path.splitext(filename)[0]
        return filename_without_extension
    
    def read_file_content(self):
        try:
            with open(self.filepath, "r") as file:
                content = file.read()
                # Remove comments before validating syntax
                content = re.sub(r"#.*", "", content)
                return content
        except FileNotFoundError:
            raise FileNotFoundError(f"{self.filename}.lxe not found.")
    
    def validate_syntax(self, content):
        try:
            self.validate_import_statement(content)
            self.validate_export_syntax(content)
            self.validate_builder_function(content)
            self.validate_math_import(content)
        except (ImportError, SyntaxError) as e:
            line_number = content.count("\n", 0, content.find(e.args[0])) + 1
            raise type(e)(f"{e.args[0]} at line {line_number} in {self.filename}.lxe")
    
    def validate_import_statement(self, content):
        import_pattern = r"import\s+{\s*enviroment\s*}\s+from\s+'LxEnviroment\.env'"
        if not re.search(import_pattern, content):
            raise ImportError("ImportError: 'enviroment' is not imported from 'LxEnviroment.env'.")

    def validate_export_syntax(self, content):
        export_pattern = r"enviroment\.export\s*=>\s*\(builder\)\s*=\s*{[^{}]*}"
        if not re.search(export_pattern, content):
            raise SyntaxError("SyntaxError: Unable to find valid 'enviroment.export => (builder) = {}' syntax.")

    def validate_builder_function(self, content):
        builder_pattern = r"=>\s*\(builder\)\s*=\s*{[^{}]*}"
        if not re.search(builder_pattern, content):
            raise SyntaxError("SyntaxError: 'builder' function is missing.")

    def validate_math_import(self, content):
        math_modules = ['add', 'sub', 'div', 'mult']
        for module in math_modules:
            import_pattern = fr"import\s+{{\s*{module}\s*}}\s+from\s+'LxEnviroment\.mathematic'"
            if re.search(import_pattern, content):
                return module
        raise ImportError("ImportError: Math modules not imported from 'LxEnviroment.mathematic'.")

    def perform_operation(self, operation, values):
        if operation == 'add':
            return sum(values)
        elif operation == 'sub':
            return values[0] - sum(values[1:])
        elif operation == 'mult':
            result = 1
            for val in values:
                result *= val
            return result
        elif operation == 'div':
            result = values[0]
            for val in values[1:]:
                result /= val
            return result

    def execute_math_operation(self, content):
        math_module = self.validate_math_import(content)
        pattern = fr"{math_module}\(\)\s*->\s*\(([^()]*)\)\s*->\s*{{type\s+either\s+(int|float|str)}}"
        match = re.search(pattern, content)
        if match:
            values = [eval(val.strip()) for val in match.group(1).split(',')]
            result = self.perform_operation(math_module, values)
            return result
        else:
            raise SyntaxError(f"SyntaxError: Invalid syntax for {math_module} operation.")

    def extract_environment_variables(self, content):
        self.validate_syntax(content)

        pattern = r"enviroment\.export\s*=>\s*\(builder\)\s*=\s*({[^{}]*})"
        match = re.search(pattern, content, re.DOTALL)
        if match:
            return match.group(1)
        else:
            raise SyntaxError("SyntaxError: Unable to extract environment variables.")
    
    def parse_environment_variables(self, env_str):
        try:
            return json.loads(env_str)
        except json.JSONDecodeError:
            raise ValueError("ValueError: Invalid JSON format for environment variables.")
    
    def get_environment_variables(self):
        content = self.read_file_content()
        env_content = self.extract_environment_variables(content)
        return self.parse_environment_variables(env_content)
    
    def set_environment_variable(self, key, value):
        content = self.read_file_content()
        env_content = self.extract_environment_variables(content)
        env_dict = self.parse_environment_variables(env_content)
        
        env_dict[key] = value
        
        with open(self.filepath, "w") as file:
            content = re.sub(r"(enviroment\.export\s*=>\s*\(builder\)\s*=\s*)({[^{}]*})", f"\\1{json.dumps(env_dict, indent=4)}", content, flags=re.DOTALL)
            file.write(content)
    
    def create_lxe_file(self):
        try:
            with open(self.filepath, "w") as file:
                file.write('import { enviroment } from \'LxEnviroment.env\'\n\n')
                file.write('enviroment.export => (builder) = {}\n')
        except IOError:
            raise IOError(f"Error: Failed to create {self.filename}.lxe file.")


import logging
from aquilify.settings.base import settings
from aquilify.utils.module_loading import import_string
from aquilify.exception.base import ImproperlyConfigured

class DatabaseVendor:
    SQLITE = "aquilify.orm.Sqlite3"

class UnsupportedVendor(ImproperlyConfigured):
    pass

class DatabaseConnectionError(ImproperlyConfigured):
    pass

class DatabaseConnectionManager:
    def __init__(self, default_vendor: str = DatabaseVendor.SQLITE, default_name: str = 'db.sqlite3') -> None:
        self.default_vendor = default_vendor
        self.default_name = default_name
        self.logger = logging.getLogger(__name__)

    def _validate_database_settings(self, settings_data):
        if not settings_data or not settings_data.get("default"):
            raise ImproperlyConfigured("The 'DATABASE' settings are missing or improperly configured.")
        engine = settings_data['default'].get("ENGINE", self.default_vendor)
        if engine != DatabaseVendor.SQLITE:
            raise UnsupportedVendor(f"Vendor '{engine}' isn't supported by aquilify by default.")

    def _get_engine_module(self, engine):
        try:
            engine_module = import_string(engine)
            return engine_module
        except ImportError as e:
            raise ImproperlyConfigured(f"Invalid engine '{engine}' or Database connection not configured properly: {e}")

    def _establish_connection(self, engine_module, name):
        try:
            db_connection = engine_module(name)
            if not db_connection:
                raise DatabaseConnectionError("Failed to establish the database connection.")
            return db_connection
        except Exception as e:
            raise DatabaseConnectionError(f"Error while establishing database connection: {e}")

    def _get_connection(self):
        try:
            database_settings = getattr(settings, 'DATABASE')
            self._validate_database_settings(database_settings)
            default_settings = database_settings.get("default", {})
            engine = default_settings.get("ENGINE", self.default_vendor)
            name = default_settings.get('NAME', self.default_name)

            engine_module = self._get_engine_module(engine)
            return self._establish_connection(engine_module, name)
        except (ImproperlyConfigured, UnsupportedVendor, DatabaseConnectionError) as e:
            self.logger.error(str(e))
            raise



import random
import time
import hashlib
import string
from typing import List, Optional

class EIDS:
    @staticmethod
    def eid1(node_id: int) -> str:
        timestamp: int = int(time.time() * 1e9)
        clock_seq: int = random.getrandbits(14)
        eid: int = (timestamp << 64) | (clock_seq << 50) | 0x1000000000000000
        eid |= node_id
        return EIDS.format_eid(eid)
    
    @staticmethod
    def eid2(node_id: int, domain_id: int) -> str:
        timestamp: int = int(time.time())
        clock_seq: int = random.getrandbits(14)
        eid: int = (timestamp << 32) | (clock_seq << 16) | 0x2000
        eid |= (node_id << 48) | domain_id
        return EIDS.format_eid(eid)
    
    @staticmethod
    def eid3(name: str) -> str:
        namespace_eid: str = EIDS.eid1(node_id=0x123456789abc)
        hashed_name: str = hashlib.md5((namespace_eid + name).encode('utf-8')).hexdigest()
        eid: str = EIDS.format_advance(hashed_name, version=3)
        return eid
    
    @staticmethod
    def eid4() -> str:
        eid: int = random.getrandbits(128)
        return EIDS.format_eid(eid)
    
    @staticmethod
    def eid5(namespace: str, name: str) -> str:
        namespace_eid: str = EIDS.eid1(node_id=0x123456789abc)
        hashed_name: str = hashlib.sha1((namespace_eid + name).encode('utf-8')).hexdigest()
        namespace_hex: str = namespace.replace('-', '')
        eid: str = hashed_name[:16] + '-' + hashed_name[16:20] + '-5' + hashed_name[21:24] + '-' + namespace_hex[:4] + '-' + namespace_hex[4:16]
        return eid
    
    @staticmethod
    def default_end() -> int:
        timestamp: int = int(time.time())
        random_part: int = random.getrandbits(64)
        eid: int = (timestamp << 64) | random_part
        return eid
    
    @staticmethod
    def format_advance(eid_str: str, version: int) -> str:
        parts: List[str] = [eid_str[:8], eid_str[8:12], str(version) + eid_str[13:16], eid_str[16:20], eid_str[20:]]
        return '-'.join(parts)
    
    @staticmethod
    def format_eid(eid: int) -> str:
        eid_str: str = f"{eid:032x}"
        formatted_eid: str = f"{eid_str[:8]}-{eid_str[8:12]}-{eid_str[12:16]}-{eid_str[16:20]}-{eid_str[20:]}"
        return formatted_eid
    
    @staticmethod
    def random(length: int = 10) -> str:
        if length < 2:
            raise ValueError("Length must be at least 2.")
        remaining_length: int = length - 2
        random_digits: str = ''.join(random.choice(string.digits) for _ in range(remaining_length))
        unique_id: str = random_digits
        return unique_id

class Eid:
    def __init__(self, node: Optional[int] = None, clock_seq: Optional[int] = None) -> None:
        self.node: int = node if node is not None else random.getrandbits(48)
        self.clock_seq: int = clock_seq if clock_seq is not None else random.getrandbits(14)
        self.variant_bits: int = 0b1000

    def chip1(self) -> str:
        timestamp: int = int(time.time() * 1e7) + 0x01b21dd213814000
        timestamp_hex: str = format(timestamp, '032x')
        clock_seq_hex: str = format(self.clock_seq, '04x')
        node_hex: str = format(self.node, '012x')
        eid: str = f"{timestamp_hex[:8]}-{timestamp_hex[8:12]}-{timestamp_hex[12:16]}-{clock_seq_hex[0]}{timestamp_hex[16:]}-{node_hex}"
        return eid

    def chip3(self, name: str, namespace_eid: str) -> str:
        hash_obj: hashlib._hashlib.HASH = hashlib.sha1(namespace_eid.encode('utf-8') + name.encode('utf-8'))
        hash_bytes: bytes = hash_obj.digest()
        hash_int: int = int.from_bytes(hash_bytes, 'big')
        eid: str = f"{hash_int:032x}"
        return eid

    def chip4(self) -> str:
        random_bits_hex: str = format(random.getrandbits(128), '032x')
        eid: str = f"{random_bits_hex[:8]}-{random_bits_hex[8:12]}-4{random_bits_hex[13:16]}-{self.variant_bits}{random_bits_hex[16:20]}-{random_bits_hex[20:]}"
        return eid

    def chip2(self, domain: Optional[str] = None) -> str:
        if domain is None:
            raise ValueError("Domain is required for version 2 UUID.")
        timestamp: int = int(time.time() * 1e7) + 0x01b21dd213814000
        timestamp_hex: str = format(timestamp, '032x')
        clock_seq_hex: str = format(self.clock_seq, '04x')
        node_hex: str = format(self.node, '012x')
        eid: str = f"{domain[:8]}-{timestamp_hex[8:12]}-{timestamp_hex[12:16]}-{clock_seq_hex[0]}{timestamp_hex[16:]}-{node_hex}"
        return eid

eid = EIDS()



import os
import platform

def get_environment_variables():
    return {
        var: os.getenv(var)
        for var in [
            'ALLUSERSPROFILE', 'APPDATA', 'CHROME_CRASHPAD_PIPE_NAME', 'COLORTERM', 'COMMONPROGRAMFILES',
            'COMMONPROGRAMFILES(X86)', 'COMMONPROGRAMW6432', 'COMPUTERNAME', 'COMSPEC', 'DRIVERDATA',
            'LOCALAPPDATA', 'LOGONSERVER', 'NUMBER_OF_PROCESSORS', 'ONEDRIVE', 'ONEDRIVECONSUMER',
            'ORIGINAL_XDG_CURRENT_DESKTOP'
        ]
    }

def get_user_info():
    try:
        username = os.getlogin()
    except OSError:
        username = "Unavailable"
    return {
        'Username': username,
        'Home Directory': os.path.expanduser('~'),
    }

def get_system_paths():
    return {
        'Root Directory': os.path.abspath('/'),
        'Temp Directory': os.getenv('TEMP'),
    }

def get_sys_info():
    sys_info = {
        'System': platform.system(),
        'Node Name': platform.node(),
        'Release': platform.release(),
        'Version': platform.version(),
        'Machine': platform.machine(),
        'Processor': platform.processor(),
        **get_environment_variables(),
        **get_user_info(),
        **get_system_paths(),
    }

    return sys_info


import os
import shutil
import asyncio
from aiofiles import open as aio_open
from typing import List, Tuple, Optional

class FileSizeLimitExceeded(Exception):
    pass

class FileSystemStorage:
    def __init__(self, base_location: str) -> None:
        self.base_location = base_location

    async def save(self, name: str, content: bytes) -> None:
        file_path = os.path.join(self.base_location, name)
        async with aio_open(file_path, 'wb') as file:
            await file.write(content)

    async def open(self, name: str, mode='rb'):
        file_path = os.path.join(self.base_location, name)
        return await aio_open(file_path, mode)

    async def read_text(self, name: str) -> str:
        file_path = os.path.join(self.base_location, name)
        async with aio_open(file_path, 'r', encoding='utf-8') as file:
            return await file.read()

    async def write_text(self, name: str, content: str) -> None:
        file_path = os.path.join(self.base_location, name)
        async with aio_open(file_path, 'w', encoding='utf-8') as file:
            await file.write(content)

    async def append_text(self, name: str, content: str) -> None:
        file_path = os.path.join(self.base_location, name)
        async with aio_open(file_path, 'a', encoding='utf-8') as file:
            await file.write(content)

    async def delete(self, name: str) -> None:
        file_path = os.path.join(self.base_location, name)
        try:
            os.remove(file_path)
        except FileNotFoundError:
            pass

    async def exists(self, name: str) -> bool:
        file_path = os.path.join(self.base_location, name)
        return os.path.exists(file_path)

    async def is_file(self, name: str) -> bool:
        file_path = os.path.join(self.base_location, name)
        return os.path.isfile(file_path)

    async def is_dir(self, name: str) -> bool:
        file_path = os.path.join(self.base_location, name)
        return os.path.isdir(file_path)

    async def listdir(self, path: str = '.') -> List[str]:
        directory = os.path.join(self.base_location, path)
        if os.path.isdir(directory):
            return os.listdir(directory)
        return []

    async def rename(self, old_name: str, new_name: str) -> None:
        old_path = os.path.join(self.base_location, old_name)
        new_path = os.path.join(self.base_location, new_name)
        try:
            os.rename(old_path, new_path)
        except FileNotFoundError:
            pass

    async def move(self, old_name: str, new_name: str) -> None:
        old_path = os.path.join(self.base_location, old_name)
        new_path = os.path.join(self.base_location, new_name)
        try:
            shutil.move(old_path, new_path)
        except FileNotFoundError:
            pass

    async def copy(self, old_name: str, new_name: str) -> None:
        old_path = os.path.join(self.base_location, old_name)
        new_path = os.path.join(self.base_location, new_name)
        try:
            shutil.copy2(old_path, new_path)
        except FileNotFoundError:
            pass

    async def mkdir(self, name: str) -> None:
        directory_path = os.path.join(self.base_location, name)
        try:
            os.makedirs(directory_path)
        except FileExistsError:
            pass

    async def stat(self, name: str) -> Optional[os.stat_result]:
        file_path = os.path.join(self.base_location, name)
        try:
            return os.stat(file_path)
        except FileNotFoundError:
            pass

    async def chmod(self, name: str, mode: int) -> None:
        file_path = os.path.join(self.base_location, name)
        try:
            os.chmod(file_path, mode)
        except FileNotFoundError:
            pass

    async def rmtree(self, name: str) -> None:
        directory_path = os.path.join(self.base_location, name)
        try:
            shutil.rmtree(directory_path)
        except FileNotFoundError:
            pass

    async def walk(self, top: str = '.') -> Tuple[List[str], List[str]]:
        top_path = os.path.join(self.base_location, top)
        files = []
        dirs = []
        for root, directories, filenames in os.walk(top_path):
            for directory in directories:
                dirs.append(os.path.relpath(os.path.join(root, directory), self.base_location))
            for filename in filenames:
                files.append(os.path.relpath(os.path.join(root, filename), self.base_location))
        return dirs, files

    async def abspath(self, name: str) -> Optional[str]:
        file_path = os.path.join(self.base_location, name)
        try:
            return os.path.abspath(file_path)
        except FileNotFoundError:
            pass

    async def get_file_size(self, name: str) -> Optional[int]:
        file_path = os.path.join(self.base_location, name)
        try:
            return os.path.getsize(file_path)
        except FileNotFoundError:
            return None

    async def check_file_size(self, name: str, max_size: int) -> None:
        file_size = await self.get_file_size(name)
        if file_size and file_size > max_size:
            raise FileSizeLimitExceeded(f"File '{name}' exceeds the maximum allowed size.")

    async def change_permissions(self, name: str, mode: int) -> None:
        file_path = os.path.join(self.base_location, name)
        try:
            os.chmod(file_path, mode)
        except FileNotFoundError:
            pass

    async def remove(self, name: str) -> None:
        file_path = os.path.join(self.base_location, name)
        loop = asyncio.get_event_loop()
        try:
            await loop.run_in_executor(None, os.remove, file_path)
        except FileNotFoundError:
            pass





from datetime import datetime

class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

AQUILIFY_THEME = """
*************** TERMINAL ***************
*          WELCOME TO AQUILIFY         *
*                                      *
* DATETIME - {}      *
* VERSION - 1.14                       *
*                                      *
****************************************
""".format(
    datetime.now().strftime("%d/%m/%Y, %H:%M:%S")
)

AX_VERSION = [
    f"Aquilify v1.10 (previous) ...{Colors.WARNING}Outdated!{Colors.ENDC}",
    f"Aquilify v1.11 (previous-fixed) ...{Colors.WARNING}Outdated!{Colors.ENDC}",
    f"Aquilify v1.12 (--upgrade) ...{Colors.WARNING}Outdated!{Colors.ENDC}",
    f"Aquilify v1.13 (--upgrade) ...{Colors.OKGREEN}Stable!{Colors.ENDC}",
    f"{Colors.OKGREEN}Aquilify v1.14 (Latest){Colors.ENDC} ...{Colors.OKGREEN}Stable!{Colors.ENDC}"
]


import typing
import warnings
from os import PathLike

from aquilify.wrappers import Request
from aquilify.responses import HTMLResponse

try:
    import jinja2
    if hasattr(jinja2, "pass_context"):
        pass_context = jinja2.pass_context
    else:  # pragma: nocover
        pass_context = jinja2.contextfunction  # type: ignore[attr-defined]
except ModuleNotFoundError:  # pragma: nocover
    jinja2 = None  # type: ignore[assignment]

from aquilify.settings.templates import TemplateConfigSettings

try:
    _config_settings = TemplateConfigSettings()
    _config_settings.fetch()
    _settings = _config_settings.template_data[0]
except Exception as e:
    raise Exception("Template settings not found in settings.py, configure it before you use.")

class _TemplateResponse(HTMLResponse):
    def __init__(
        self,
        template: typing.Any,
        context: typing.Dict[str, typing.Any],
        status_code: int = 200,
        headers: typing.Optional[typing.Mapping[str, str]] = None,
        content_type: typing.Optional[str] = None
    ):
        self.template = template
        self.context = context
        content = template.render(context)
        super().__init__(content, status_code, headers, content_type)

class Jinja2Templates:
    """
    templates = Jinja2Templates("templates")

    return templates.TemplateResponse("index.html", {"request": request})
    """

    @typing.overload
    def __init__(
        self,
        directory: "typing.Union[str, PathLike[typing.AnyStr], typing.Sequence[typing.Union[str, PathLike[typing.AnyStr]]]]",  # noqa: E501
        *,
        context_processors: typing.Optional[
            typing.List[typing.Callable[[Request], typing.Dict[str, typing.Any]]]
        ] = None,
        **env_options: typing.Any,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self,
        *,
        env: "jinja2.Environment",
        context_processors: typing.Optional[
            typing.List[typing.Callable[[Request], typing.Dict[str, typing.Any]]]
        ] = None,
    ) -> None:
        ...

    def __init__(
        self,
        directory: "typing.Union[str, PathLike[typing.AnyStr], typing.Sequence[typing.Union[str, PathLike[typing.AnyStr]]], None]" = _settings.get('dirs') or None,  # noqa: E501
        *,
        context_processors: typing.Optional[
            typing.List[typing.Callable[[Request], typing.Dict[str, typing.Any]]]
        ] = _settings['options'].get('context_processors') or None,
        env: typing.Optional["jinja2.Environment"] = _settings['options'].get('enviroment') or None,
        **env_options: typing.Any,
    ) -> None:
        if env_options:
            warnings.warn(
                "Extra environment options are deprecated. Use a preconfigured jinja2.Environment instead.",  # noqa: E501
                DeprecationWarning,
            )
        assert jinja2 is not None, "jinja2 must be installed to use Jinja2Templates"
        assert directory or env, "either 'directory' or 'env' arguments must be passed"
        assert _settings.get('backend') != 'aquilify.template.jinja2', "Aquilify currently support Jinj2 as template engine"
        self.context_processors = context_processors or []
        self.autoscape =  _settings['options'].get('autoscape') or True
        self.extensions =  _settings['options'].get('extensions') or ()
        self.cache_size =  _settings['options'].get('cache_size') or 400
        if directory is not None:
            self.env = self._create_env(directory, **env_options)
        elif env is not None:
            self.env = env

    def _create_env(
        self,
        directory: "typing.Union[str, PathLike[typing.AnyStr], typing.Sequence[typing.Union[str, PathLike[typing.AnyStr]]]]",  # noqa: E501
        **env_options: typing.Any,
    ) -> "jinja2.Environment":
        loader = jinja2.FileSystemLoader(directory)
        env_options.setdefault("loader", loader)
        env_options.setdefault("autoescape", self.autoscape)
        env_options.setdefault("extensions", self.extensions)
        env_options.setdefault("cache_size", self.cache_size)

        return jinja2.Environment(**env_options)

    def get_template(self, name: str) -> "jinja2.Template":
        return self.env.get_template(name)

    @typing.overload
    async def TemplateResponse(
        self,
        request: Request,
        name: str,
        context: typing.Optional[typing.Dict[str, typing.Any]] = None,
        status_code: int = 200,
        headers: typing.Optional[typing.Mapping[str, str]] = None,
        content_type: typing.Optional[str] = None
    ) -> _TemplateResponse:
        ...

    @typing.overload
    async def TemplateResponse(
        self,
        name: str,
        context: typing.Optional[typing.Dict[str, typing.Any]] = None,
        status_code: int = 200,
        headers: typing.Optional[typing.Mapping[str, str]] = None,
        content_type: typing.Optional[str] = None
    ) -> _TemplateResponse:
        ...

    async def TemplateResponse(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> _TemplateResponse:
        if args:
            if isinstance(
                args[0], str
            ):
                warnings.warn(
                    "The `name` is not the first parameter anymore. "
                    "The first parameter should be the `Request` instance.\n"
                    'Replace `TemplateResponse(name, {"request": request})` by `TemplateResponse(request, name)`.',  # noqa: E501
                    DeprecationWarning,
                )

                name = args[0]
                context = args[1] if len(args) > 1 else kwargs.get("context", {})
                status_code = (
                    args[2] if len(args) > 2 else kwargs.get("status_code", 200)
                )
                headers = args[2] if len(args) > 2 else kwargs.get("headers")
                content_type = args[3] if len(args) > 3 else kwargs.get("content_type")

                if "request" not in context:
                    raise ValueError('context must include a "request" key')
                request = context["request"]
            else:  # the first argument is a request instance (new style)
                request = args[0]
                name = args[1] if len(args) > 1 else kwargs["name"]
                context = args[2] if len(args) > 2 else kwargs.get("context", {})
                status_code = (
                    args[3] if len(args) > 3 else kwargs.get("status_code", 200)
                )
                headers = args[4] if len(args) > 4 else kwargs.get("headers")
                content_type = args[5] if len(args) > 5 else kwargs.get("content_type")
        else:  # all arguments are kwargs
            if "request" not in kwargs:
                warnings.warn(
                    "The `TemplateResponse` now requires the `request` argument.\n"
                    'Replace `TemplateResponse(name, {"context": context})` by `TemplateResponse(request, name)`.',  # noqa: E501
                    DeprecationWarning,
                )
                if "request" not in kwargs.get("context", {}):
                    raise ValueError('context must include a "request" key')

            context = kwargs.get("context", {})
            request = kwargs.get("request", context.get("request"))
            name = typing.cast(str, kwargs["name"])
            status_code = kwargs.get("status_code", 200)
            headers = kwargs.get("headers")
            content_type = kwargs.get("content_type")

        context.setdefault("request", request)
        for context_processor in self.context_processors:
            context.update(await context_processor(context, request))

        template = self.get_template(name)
        return _TemplateResponse(
            template,
            context,
            status_code=status_code,
            headers=headers,
            content_type=content_type,
        )



import hashlib
import hmac
import secrets

from ..settings import settings
from .encoding import force_bytes


class InvalidAlgorithm(ValueError):
    pass


def salted_hmac(key_salt, value, secret=None, *, algorithm="sha1"):
    if secret is None:
        secret = settings.SECRET_KEY

    key_salt = force_bytes(key_salt)
    secret = force_bytes(secret)
    try:
        hasher = getattr(hashlib, algorithm)
    except AttributeError as e:
        raise InvalidAlgorithm(
            "%r is not an algorithm accepted by the hashlib module." % algorithm
        ) from e
    key = hasher(key_salt + secret).digest()
    return hmac.new(key, msg=force_bytes(value), digestmod=hasher)


RANDOM_STRING_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


def get_random_string(length, allowed_chars=RANDOM_STRING_CHARS):
    return "".join(secrets.choice(allowed_chars) for i in range(length))


def constant_time_compare(val1, val2):
    return secrets.compare_digest(force_bytes(val1), force_bytes(val2))


def pbkdf2(password, salt, iterations, dklen=0, digest=None):
    if digest is None:
        digest = hashlib.sha256
    dklen = dklen or None
    password = force_bytes(password)
    salt = force_bytes(salt)
    return hashlib.pbkdf2_hmac(digest().name, password, salt, iterations, dklen)


from typing import List, Dict, Optional, Tuple, Any, Callable
import logging
import zlib
import base64

from . import WebSocket

class Channel:
    def __init__(self):
        self.connected_websockets: List[WebSocket] = []
        self.subscribed_channels: Dict[str, List[WebSocket]] = {}
        self.groups: Dict[str, List[WebSocket]] = {}
        self.message_history: Dict[WebSocket, List[str]] = {}
        self.message_queue: Dict[str, List[Tuple[str, str]]] = {}
        self.authenticated_websockets: Dict[WebSocket, Any] = {}
        self.event_listeners: Dict[str, List[Callable[[Any], Any]]] = {}
        self.logger = logging.getLogger(__name__)

    async def connect(self, websocket: WebSocket, user: Optional[Any] = None):
        await websocket.accept()
        self.connected_websockets.append(websocket)
        self.message_history[websocket] = []
        self.authenticated_websockets[websocket] = user

    async def disconnect(self, websocket: WebSocket):
        await websocket.close()
        self.connected_websockets.remove(websocket)
        self._remove_from_subscriptions(websocket)
        del self.message_history[websocket]
        self.authenticated_websockets.pop(websocket, None)

    async def subscribe(self, websocket: WebSocket, channel_name: str):
        self.subscribed_channels.setdefault(channel_name, []).append(websocket)
        if channel_name in self.message_history:
            for message in self.message_history[websocket]:
                await websocket.send_text(message)
        if channel_name in self.message_queue:
            for sender, message in self.message_queue[channel_name]:
                await websocket.send_text(f"[{sender}] {message}")

    async def unsubscribe(self, websocket: WebSocket, channel_name: str):
        if channel_name in self.subscribed_channels:
            self.subscribed_channels[channel_name].remove(websocket)

    async def group_add(self, group_name: str, websocket: WebSocket):
        self.groups.setdefault(group_name, []).append(websocket)

    async def group_discard(self, group_name: str, websocket: WebSocket):
        if group_name in self.groups:
            self.groups[group_name].remove(websocket)

    async def group_send(self, group_name: str, message: str, sender: Optional[str] = None):
        if group_name in self.groups:
            for websocket in self.groups[group_name]:
                await websocket.send_text(message)

    async def send_to_user(self, user: Any, message: str):
        for websocket, auth_user in self.authenticated_websockets.items():
            if auth_user == user:
                await websocket.send_text(message)

    async def broadcast(self, message: str, sender: Optional[str] = None, compression: bool = False):
        try:
            for websocket in self.connected_websockets:
                await self._send_message(message, websocket, compression)
        except Exception as e:
            self.logger.error(f"Error broadcasting message: {str(e)}")

    async def broadcast_to_channel(
        self, message: str, channel_name: str, sender: Optional[str] = None, compression: bool = False
    ):
        try:
            if channel_name in self.subscribed_channels:
                for websocket in self.subscribed_channels[channel_name]:
                    await self._send_message(message, websocket, compression)
        except Exception as e:
            self.logger.error(f"Error broadcasting to channel {channel_name}: {str(e)}")

    async def direct_message(self, message: str, recipient: WebSocket, sender: str):
        try:
            await recipient.send_text(f"[DM from {sender}] {message}")
        except Exception as e:
            self.logger.error(f"Error sending direct message: {str(e)}")

    def queue_message(self, sender: str, message: str, channel_name: str):
        self.message_queue.setdefault(channel_name, []).append((sender, message))

    def _remove_from_subscriptions(self, websocket: WebSocket):
        for ws_list in self.subscribed_channels.values():
            if websocket in ws_list:
                ws_list.remove(websocket)

    async def _send_message(self, message: str, websocket: WebSocket, compression: bool):
        try:
            if compression:
                compressed_message = zlib.compress(message.encode('utf-8'))
                compressed_message = base64.b64encode(compressed_message).decode('utf-8')
                await websocket.send_text(compressed_message)
            else:
                await websocket.send_text(message)
            self.message_history[websocket].append(message)
        except Exception as e:
            self.logger.error(f"Error sending message: {str(e)}")

    def add_event_listener(self, event_name: str, handler: Callable[[Any], Any]):
        self.event_listeners.setdefault(event_name, []).append(handler)

    async def emit_event(self, event_name: str, data: Any):
        if event_name in self.event_listeners:
            for handler in self.event_listeners[event_name]:
                try:
                    await handler(data)
                except Exception as e:
                    self.logger.error(f"Error handling event {event_name}: {str(e)}")



import importlib.util

class TrustedHostConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_trustedhost_config(self):
        try:
            settings = self._load_settings_module()
            allowed_hosts = getattr(settings, 'ALLOWED_HOSTS', ['*'])
            allow_subdomains = getattr(settings, 'ALLOW_SUBDOMAINS', True)
            redirect_on_fails = getattr(settings, 'REDIRECT_ON_FAILS', False)
            redirect_url = getattr(settings, 'REDIRECT_URL', None)
            enforce_https = getattr(settings, 'ENFORCE_HTTPS', False)
            www_redirect = getattr(settings, 'WWW_REDIRECT', False)

            return { 'allowed_hosts': allowed_hosts, 'allow_subdomains': allow_subdomains, 'redirect_on_fails': redirect_on_fails,
                'redirect_url': redirect_url, "enforce_https": enforce_https, "www_redirect": www_redirect }
                    
        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings

    def fetch(self):
        data = self._fetch_trustedhost_config()
        return data


import copy
import os
import sys
from importlib import import_module
from importlib.util import find_spec as importlib_find

def cached_import(module_path, class_name):
    if not ((module := sys.modules.get(module_path))
            and (spec := getattr(module, "__spec__", None))
            and getattr(spec, "_initializing", False) is False):
        module = import_module(module_path)
    return getattr(module, class_name)

def import_string(dotted_path):
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    try:
        return cached_import(module_path, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err

def module_has_submodule(package, module_name):
    try:
        package_name = package.__name__
        package_path = package.__path__
    except AttributeError:
        return False

    full_module_name = package_name + "." + module_name
    try:
        return importlib_find(full_module_name, package_path) is not None
    except ModuleNotFoundError:
        return False

def module_dir(module):
    paths = list(getattr(module, "__path__", []))
    if len(paths) == 1:
        return paths[0]
    else:
        filename = getattr(module, "__file__", None)
        if filename is not None:
            return os.path.dirname(filename)
    raise ValueError("Cannot determine directory containing %s" % module)



import mimetypes
from email import charset as Charset
from email import encoders as Encoders
from email import generator, message_from_string
from email.errors import HeaderParseError
from email.header import Header
from email.headerregistry import Address, parser
from email.message import Message
from email.mime.base import MIMEBase
from email.mime.message import MIMEMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr, formatdate, getaddresses, make_msgid
from io import BytesIO, StringIO
from pathlib import Path

from aquilify.settings import settings
from .utils import DNS_NAME
from aquilify.utils.encoding import force_str, punycode

utf8_charset = Charset.Charset("utf-8")
utf8_charset.body_encoding = None
utf8_charset_qp = Charset.Charset("utf-8")
utf8_charset_qp.body_encoding = Charset.QP

DEFAULT_ATTACHMENT_MIME_TYPE = "application/octet-stream"
RFC5322_EMAIL_LINE_LENGTH_LIMIT = 998

class BadHeaderError(ValueError):
    pass

ADDRESS_HEADERS = {
    "from", "sender", "reply-to", "to", "cc", "bcc", "resent-from",
    "resent-sender", "resent-to", "resent-cc", "resent-bcc"
}

def forbid_multi_line_headers(name, val, encoding):
    encoding = encoding or settings.DEFAULT_CHARSET
    val = str(val)
    if "\n" in val or "\r" in val:
        raise BadHeaderError(
            "Header values can't contain newlines (got %r for header %r)" % (val, name)
        )
    try:
        val.encode("ascii")
    except UnicodeEncodeError:
        if name.lower() in ADDRESS_HEADERS:
            val = ", ".join(
                sanitize_address(addr, encoding) for addr in getaddresses((val,))
            )
        else:
            val = Header(val, encoding).encode()
    else:
        if name.lower() == "subject":
            val = Header(val).encode()
    return name, val

def sanitize_address(addr, encoding):
    address = None
    if not isinstance(addr, tuple):
        addr = force_str(addr)
        try:
            token, rest = parser.get_mailbox(addr)
        except (HeaderParseError, ValueError, IndexError):
            raise ValueError('Invalid address "%s"' % addr)
        else:
            if rest:
                raise ValueError(
                    'Invalid address; only %s could be parsed from "%s"' % (token, addr)
                )
            nm = token.display_name or ""
            localpart = token.local_part
            domain = token.domain or ""
    else:
        nm, address = addr
        if "@" not in address:
            raise ValueError(f'Invalid address "{address}"')
        localpart, domain = address.rsplit("@", 1)

    address_parts = nm + localpart + domain
    if "\n" in address_parts or "\r" in address_parts:
        raise ValueError("Invalid address; address parts cannot contain newlines.")

    try:
        nm.encode("ascii")
        nm = Header(nm).encode()
    except UnicodeEncodeError:
        nm = Header(nm, encoding).encode()
    try:
        localpart.encode("ascii")
    except UnicodeEncodeError:
        localpart = Header(localpart, encoding).encode()
    domain = punycode(domain)

    parsed_address = Address(username=localpart, domain=domain)
    return formataddr((nm, parsed_address.addr_spec))

class MIMEMixin:
    def as_string(self, unixfrom=False, linesep="\n"):
        fp = StringIO()
        g = generator.Generator(fp, mangle_from_=False)
        g.flatten(self, unixfrom=unixfrom, linesep=linesep)
        return fp.getvalue()

    def as_bytes(self, unixfrom=False, linesep="\n"):
        fp = BytesIO()
        g = generator.BytesGenerator(fp, mangle_from_=False)
        g.flatten(self, unixfrom=unixfrom, linesep=linesep)
        return fp.getvalue()

class SafeMIMEMessage(MIMEMixin, MIMEMessage):
    def __setitem__(self, name, val):
        name, val = forbid_multi_line_headers(name, val, "ascii")
        MIMEMessage.__setitem__(self, name, val)

class SafeMIMEText(MIMEMixin, MIMEText):
    def __init__(self, _text, _subtype="plain", _charset=None):
        self.encoding = _charset
        MIMEText.__init__(self, _text, _subtype=_subtype, _charset=_charset)

    def __setitem__(self, name, val):
        name, val = forbid_multi_line_headers(name, val, self.encoding)
        MIMEText.__setitem__(self, name, val)

    def set_payload(self, payload, charset=None):
        if charset == "utf-8" and not isinstance(charset, Charset.Charset):
            has_long_lines = any(
                len(line.encode()) > RFC5322_EMAIL_LINE_LENGTH_LIMIT
                for line in payload.splitlines()
            )
            charset = utf8_charset_qp if has_long_lines else utf8_charset
        MIMEText.set_payload(self, payload, charset=charset)

class SafeMIMEMultipart(MIMEMixin, MIMEMultipart):
    def __init__(
        self, _subtype="mixed", boundary=None, _subparts=None, encoding=None, **_params
    ):
        self.encoding = encoding
        MIMEMultipart.__init__(self, _subtype, boundary, _subparts, **_params)

    def __setitem__(self, name, val):
        name, val = forbid_multi_line_headers(name, val, self.encoding)
        MIMEMultipart.__setitem__(self, name, val)

class EmailMessage:
    content_subtype = "plain"
    mixed_subtype = "mixed"
    encoding = None

    def __init__(
        self,
        subject="",
        body="",
        from_email=None,
        to=None,
        bcc=None,
        connection=None,
        attachments=None,
        headers=None,
        cc=None,
        reply_to=None,
    ):
        if to:
            if isinstance(to, str):
                raise TypeError('"to" argument must be a list or tuple')
            self.to = list(to)
        else:
            self.to = []
        if cc:
            if isinstance(cc, str):
                raise TypeError('"cc" argument must be a list or tuple')
            self.cc = list(cc)
        else:
            self.cc = []
        if bcc:
            if isinstance(bcc, str):
                raise TypeError('"bcc" argument must be a list or tuple')
            self.bcc = list(bcc)
        else:
            self.bcc = []
        if reply_to:
            if isinstance(reply_to, str):
                raise TypeError('"reply_to" argument must be a list or tuple')
            self.reply_to = list(reply_to)
        else:
            self.reply_to = []
        self.from_email = from_email or settings.DEFAULT_FROM_EMAIL
        self.subject = subject
        self.body = body or ""
        self.attachments = []
        if attachments:
            for attachment in attachments:
                if isinstance(attachment, MIMEBase):
                    self.attach(attachment)
                else:
                    self.attach(*attachment)
        self.extra_headers = headers or {}
        self.connection = connection

    def get_connection(self, fail_silently=False):
        from . import get_connection

        if not self.connection:
            self.connection = get_connection(fail_silently=fail_silently)
        return self.connection

    def message(self):
        encoding = self.encoding or settings.DEFAULT_CHARSET
        msg = SafeMIMEText(self.body, self.content_subtype, encoding)
        msg = self._create_message(msg)
        msg["Subject"] = self.subject
        msg["From"] = self.extra_headers.get("From", self.from_email)
        self._set_list_header_if_not_empty(msg, "To", self.to)
        self._set_list_header_if_not_empty(msg, "Cc", self.cc)
        self._set_list_header_if_not_empty(msg, "Reply-To", self.reply_to)
        header_names = [key.lower() for key in self.extra_headers]
        if "date" not in header_names:
            msg["Date"] = formatdate(localtime=settings.EMAIL_USE_LOCALTIME)
        if "message-id" not in header_names:
            msg["Message-ID"] = make_msgid(domain=DNS_NAME)
        for name, value in self.extra_headers.items():
            if name.lower() != "from":
                msg[name] = value
        return msg

    def recipients(self):
        return [email for email in (self.to + self.cc + self.bcc) if email]

    def send(self, fail_silently=False):
        if not self.recipients():
            return 0
        return self.get_connection(fail_silently).send_messages([self])

    def attach(self, filename=None, content=None, mimetype=None):
        if isinstance(filename, MIMEBase):
            if content is not None or mimetype is not None:
                raise ValueError(
                    "content and mimetype must not be given when a MIMEBase "
                    "instance is provided."
                )
            self.attachments.append(filename)
        elif content is None:
            raise ValueError("content must be provided.")
        else:
            mimetype = (
                mimetype
                or mimetypes.guess_type(filename)[0]
                or DEFAULT_ATTACHMENT_MIME_TYPE
            )
            basetype, subtype = mimetype.split("/", 1)

            if basetype == "text":
                if isinstance(content, bytes):
                    try:
                        content = content.decode()
                    except UnicodeDecodeError:
                        mimetype = DEFAULT_ATTACHMENT_MIME_TYPE

            self.attachments.append((filename, content, mimetype))

    def attach_file(self, path, mimetype=None):
        path = Path(path)
        with path.open("rb") as file:
            content = file.read()
            self.attach(path.name, content, mimetype)

    def _create_message(self, msg):
        return self._create_attachments(msg)

    def _create_attachments(self, msg):
        if self.attachments:
            encoding = self.encoding or settings.DEFAULT_CHARSET
            body_msg = msg
            msg = SafeMIMEMultipart(_subtype=self.mixed_subtype, encoding=encoding)
            if self.body or body_msg.is_multipart():
                msg.attach(body_msg)
            for attachment in self.attachments:
                if isinstance(attachment, MIMEBase):
                    msg.attach(attachment)
                else:
                    msg.attach(self._create_attachment(*attachment))
        return msg

    def _create_mime_attachment(self, content, mimetype):
        basetype, subtype = mimetype.split("/", 1)
        if basetype == "text":
            encoding = self.encoding or settings.DEFAULT_CHARSET
            attachment = SafeMIMEText(content, subtype, encoding)
        elif basetype == "message" and subtype == "rfc822":
            if isinstance(content, EmailMessage):
                content = content.message()
            elif not isinstance(content, Message):
                content = message_from_string(force_str(content))

            attachment = SafeMIMEMessage(content, subtype)
        else:
            attachment = MIMEBase(basetype, subtype)
            attachment.set_payload(content)
            Encoders.encode_base64(attachment)
        return attachment

    def _create_attachment(self, filename, content, mimetype=None):
        attachment = self._create_mime_attachment(content, mimetype)
        if filename:
            try:
                filename.encode("ascii")
            except UnicodeEncodeError:
                filename = ("utf-8", "", filename)
            attachment.add_header(
                "Content-Disposition", "attachment", filename=filename
            )
        return attachment

    def _set_list_header_if_not_empty(self, msg, header, values):
        if values:
            try:
                value = self.extra_headers[header]
            except KeyError:
                value = ", ".join(str(v) for v in values)
            msg[header] = value

class EmailMultiAlternatives(EmailMessage):
    alternative_subtype = "alternative"

    def __init__(
        self,
        subject="",
        body="",
        from_email=None,
        to=None,
        bcc=None,
        connection=None,
        attachments=None,
        headers=None,
        alternatives=None,
        cc=None,
        reply_to=None,
    ):
        super().__init__(
            subject,
            body,
            from_email,
            to,
            bcc,
            connection,
            attachments,
            headers,
            cc,
            reply_to,
        )
        self.alternatives = alternatives or []

    def attach_alternative(self, content, mimetype):
        if content is None or mimetype is None:
            raise ValueError("Both content and mimetype must be provided.")
        self.alternatives.append((content, mimetype))

    def _create_message(self, msg):
        return self._create_attachments(self._create_alternatives(msg))

    def _create_alternatives(self, msg):
        encoding = self.encoding or settings.DEFAULT_CHARSET
        if self.alternatives:
            body_msg = msg
            msg = SafeMIMEMultipart(
                _subtype=self.alternative_subtype, encoding=encoding
            )
            if self.body:
                msg.attach(body_msg)
            for alternative in self.alternatives:
                msg.attach(self._create_mime_attachment(*alternative))
        return msg



from typing import Any, Dict, Optional, Callable
from markupsafe import Markup
from html import escape
from ...settings.xsrf_view import XSRFConfigSettings

try:
    _config_settings = XSRFConfigSettings()
    _config_settings.fetch()
    _settings = _config_settings.template_data[0]
except Exception as e:
    raise Exception("Template settings not found in settings.py, configure it before you use.")

class XSRFContextView:
    def __init__(self) -> None:
        self.csrf: Optional[Any] = _settings.get('csrf')

    async def __call__(self, context: Dict[str, Any], request: Any) -> Dict[str, Any]:
        token: Optional[str] = await self._get_csrf_token(request) if self.csrf else None
        csrf_protect: Callable[[], Markup] = self._get_csrf_protect(token)
        context['csrf_protect'] = csrf_protect
        request.context['_csrf_token'] = token
        request.context['_csrf_view'] = self.csrf
        return context

    async def _get_csrf_token(self, request: Any) -> Optional[str]:
        if self.csrf:
            return await self.csrf.generate_csrf_token(request.remote_addr)
        return None

    def _get_csrf_protect(self, token: Optional[str]) -> Callable[[], Markup]:
        if token:
            token_name: str = self.csrf.csrf_token_key
            return lambda: Markup(f'<input name="{token_name}" type="hidden" value="{escape(token)}"></input>')
        return lambda: Markup('')



from typing import Any, Dict
from .api import get_messages
from .constants import DEFAULT_LEVELS

class Messages:
    """
    Middleware class responsible for handling messages in the application context.
    """

    async def __call__(self, context: Dict[str, Any], request) -> Dict[str, Any]:
        """
        Assigns messages and default levels to the context based on the incoming request.

        Args:
        - context (Dict[str, Any]): The application context.
        - request (Request): The incoming request.

        Returns:
        - Dict[str, Any]: The updated application context containing messages and default levels.
        """
        context['messages'] = get_messages(request)
        context['DEFAULT_MESSAGE_LEVELS'] = DEFAULT_LEVELS
        
        return context



import importlib.util

class Settings:
    def __init__(self, settings_module_path="./settings.py") -> None:
        self.settings_module_path = settings_module_path
        self._load_settings_config()

    def _load_settings_config(self):
        try:
            settings = self._load_settings_module()
            self._set_attributes(settings)

        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings

    def _set_attributes(self, settings):
        for attr_name in dir(settings):
            if not attr_name.startswith("__"):
                try:
                    setattr(self, attr_name, getattr(settings, attr_name))
                except AttributeError as e:
                    print(f"Error setting attribute {attr_name}: {e}")

settings = Settings()


# T@noql __fix:: 3001 
# Do not use this middleware this is only for development usage!

# refer :: --> @noql 3001 

import time
from typing import Callable, Awaitable, TypeVar

T = TypeVar("T")

class AquilifyProfiler:
    def __init__(self):
        self.profiles = {}

    def profile(self, handler: Callable[..., Awaitable[T]]):
        async def wrapper(request):
            start_time = time.time()
            response = await handler(request)
            end_time = time.time()
            execution_time = end_time - start_time
            self.profiles[handler.__name__] = execution_time
            return response

        return wrapper

    def print_profiles(self):
        print("Profile results:")
        for handler_name, execution_time in self.profiles.items():
            print(f"{handler_name}: {execution_time} seconds")



from ..wrappers import Request, Response
from ..settings import settings

def write_hsts_header_value(max_age: int, include_subdomains: bool) -> str:
    value = f"max-age={max_age};"

    if include_subdomains:
        value = value + " includeSubDomains;"

    return value


class HSTSMiddleware:
    """
    Middleware configuring "Strict-Transport-Security" header on responses.
    By default, it uses "max-age=31536000; includeSubDomains;".
    """

    def __init__(
        self,
        max_age: int = getattr(settings, 'HSTS_MAX_AGE', 31536000),
        include_subdomains: bool = getattr(settings, 'HSTS_INCLUDE_SUBDOMAINS', True),
    ) -> None:
        self._value = write_hsts_header_value(max_age, include_subdomains)

    async def __call__(self, request: Request, response: Response):
        response.headers["Strict-Transport-Security"] = self._value
        return response



import copy
from collections.abc import Mapping


class OrderedSet:

    def __init__(self, iterable=None):
        self.dict = dict.fromkeys(iterable or ())

    def add(self, item):
        self.dict[item] = None

    def remove(self, item):
        del self.dict[item]

    def discard(self, item):
        try:
            self.remove(item)
        except KeyError:
            pass

    def __iter__(self):
        return iter(self.dict)

    def __reversed__(self):
        return reversed(self.dict)

    def __contains__(self, item):
        return item in self.dict

    def __bool__(self):
        return bool(self.dict)

    def __len__(self):
        return len(self.dict)

    def __repr__(self):
        data = repr(list(self.dict)) if self.dict else ""
        return f"{self.__class__.__qualname__}({data})"


class MultiValueDictKeyError(KeyError):
    pass


class MultiValueDict(dict):
    def __init__(self, key_to_list_mapping=()):
        super().__init__(key_to_list_mapping)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, super().__repr__())

    def __getitem__(self, key):
        """
        Return the last data value for this key, or [] if it's an empty list;
        raise KeyError if not found.
        """
        try:
            list_ = super().__getitem__(key)
        except KeyError:
            raise MultiValueDictKeyError(key)
        try:
            return list_[-1]
        except IndexError:
            return []

    def __setitem__(self, key, value):
        super().__setitem__(key, [value])

    def __copy__(self):
        return self.__class__([(k, v[:]) for k, v in self.lists()])

    def __deepcopy__(self, memo):
        result = self.__class__()
        memo[id(self)] = result
        for key, value in dict.items(self):
            dict.__setitem__(
                result, copy.deepcopy(key, memo), copy.deepcopy(value, memo)
            )
        return result

    def __getstate__(self):
        return {**self.__dict__, "_data": {k: self._getlist(k) for k in self}}

    def __setstate__(self, obj_dict):
        data = obj_dict.pop("_data", {})
        for k, v in data.items():
            self.setlist(k, v)
        self.__dict__.update(obj_dict)

    def get(self, key, default=None):
        try:
            val = self[key]
        except KeyError:
            return default
        if val == []:
            return default
        return val

    def _getlist(self, key, default=None, force_list=False):
        try:
            values = super().__getitem__(key)
        except KeyError:
            if default is None:
                return []
            return default
        else:
            if force_list:
                values = list(values) if values is not None else None
            return values

    def getlist(self, key, default=None):
        return self._getlist(key, default, force_list=True)

    def setlist(self, key, list_):
        super().__setitem__(key, list_)

    def setdefault(self, key, default=None):
        if key not in self:
            self[key] = default
        return self[key]

    def setlistdefault(self, key, default_list=None):
        if key not in self:
            if default_list is None:
                default_list = []
            self.setlist(key, default_list)
        return self._getlist(key)

    def appendlist(self, key, value):
        self.setlistdefault(key).append(value)

    def items(self):
        for key in self:
            yield key, self[key]

    def lists(self):
        return iter(super().items())

    def values(self):
        for key in self:
            yield self[key]

    def copy(self):
        return copy.copy(self)

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError("update expected at most 1 argument, got %d" % len(args))
        if args:
            arg = args[0]
            if isinstance(arg, MultiValueDict):
                for key, value_list in arg.lists():
                    self.setlistdefault(key).extend(value_list)
            else:
                if isinstance(arg, Mapping):
                    arg = arg.items()
                for key, value in arg:
                    self.setlistdefault(key).append(value)
        for key, value in kwargs.items():
            self.setlistdefault(key).append(value)

    def dict(self):
        return {key: self[key] for key in self}


class ImmutableList(tuple):
    def __new__(cls, *args, warning="ImmutableList object is immutable.", **kwargs):
        self = tuple.__new__(cls, *args, **kwargs)
        self.warning = warning
        return self

    def complain(self, *args, **kwargs):
        raise AttributeError(self.warning)

    # All list mutation functions complain.
    __delitem__ = complain
    __delslice__ = complain
    __iadd__ = complain
    __imul__ = complain
    __setitem__ = complain
    __setslice__ = complain
    append = complain
    extend = complain
    insert = complain
    pop = complain
    remove = complain
    sort = complain
    reverse = complain


class DictWrapper(dict):
    def __init__(self, data, func, prefix):
        super().__init__(data)
        self.func = func
        self.prefix = prefix

    def __getitem__(self, key):
        use_func = key.startswith(self.prefix)
        key = key.removeprefix(self.prefix)
        value = super().__getitem__(key)
        if use_func:
            return self.func(value)
        return value


class CaseInsensitiveMapping(Mapping):
    def __init__(self, data):
        self._store = {k.lower(): (k, v) for k, v in self._unpack_items(data)}

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __len__(self):
        return len(self._store)

    def __eq__(self, other):
        return isinstance(other, Mapping) and {
            k.lower(): v for k, v in self.items()
        } == {k.lower(): v for k, v in other.items()}

    def __iter__(self):
        return (original_key for original_key, value in self._store.values())

    def __repr__(self):
        return repr({key: value for key, value in self._store.values()})

    def copy(self):
        return self

    @staticmethod
    def _unpack_items(data):
        if isinstance(data, (dict, Mapping)):
            yield from data.items()
            return
        for i, elem in enumerate(data):
            if len(elem) != 2:
                raise ValueError(
                    "dictionary update sequence element #{} has length {}; "
                    "2 is required.".format(i, len(elem))
                )
            if not isinstance(elem[0], str):
                raise ValueError(
                    "Element key %r invalid, only strings are allowed" % elem[0]
                )
            yield elem



SETTINGS = """############################################# AQUILIFY APP SETTINGS #############################################
import os
import pathlib 

"use development" # calling development server from aquilif.core.server @noql -> 5391 
# change it to "use production" in production usage.

BASE_DIR = pathlib.Path(__file__).resolve().parent

# Warning - Do not make changes to this ENTRY_POINT it may broke the whole application.
# For more info about ENTERY_POINT :: refer @noql -> `/__root__.py`

ENTRY_POINT = "__root__.__instance__"

ASGI_APPLICATION = "asgi.application"

SECRET_KEY = "%s" # replace with you own strong and secure secret key.
# A secret key for secure session data and other cryptographic operations. 
# Keep this key secret and unique for your application.

DEFAULT_CHARSET = 'UTF-8'

DEBUG = True # Disable the DEBUG mode in production usage, due to security measures.

STRICT_SLASHES = True # Strict slashes is not implemented yet, so use the regular method to add STRICT_SLASHES

METHODS_RANGE = ["GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH", "TRACE"] # # Methods Range is not implemented yet, so it uses the deafult configuration.

ROOT_ROUTING = "routing"

### Trusted Host Middleware Configuration...

ALLOWED_HOSTS = ["*"] # * indicate to allow all the hosts.

# Additional Configuration...

ALLOW_SUBDOMAINS = True
REDIRECT_ON_FAILS = False
REDIRECT_URL = ""
ENFORCE_HTTPS = False
WWW_REDIRECT = False

API_CONSOLE_URL = '/api/console'

STATIC_CHUNK_SIZE = 65536

STATIC_URL = '/static/'

STATIC_MULTI_URL = [ ] 

STATICFILE_DIRS = [
    os.path.join(BASE_DIR, 'static')
]

MEDIA_DIRS = [
    os.path.join(BASE_DIR, 'media'),
]

MEDIA_URL = "/media/"

### X-FRAME-OPTIONS Configuration...

# Replace the value according to your needs; it may conflict with the XFrameOptionsMiddleware.
# For more information, visit: http://aquilify.vvfin.in/middlewares/core/x-frame-options
# Refer to @noql -> 5369
# Learn more about security headers at: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Frame-Options

X_FAME_OPTIONS = "SAMEORIGIN"

# APPLICATION_DISPATCHER has not been confiured yet so use the manual setting to configure it.

APPLICATION_DISPATCHER = [ ]

### Templates Configuration...

# For detailed configuration information, visit: http://aquilify.vvfin.in/template/jinja2/settings
# Additional details available at @noql -> 5370.

TEMPLATES = [
    {
        "BACKEND": "aquilify.template.jinja2",
        "DIRS": [ os.path.join(BASE_DIR, "templates") ],
        "CSRF": "aquilify.security.csrf.CSRF",
        "OPTIONS": {
            "autoscape": True,
            "context_processors": [
                "aquilify.template.context_processors.URLContextProcessor",
                "aquilify.template.context_processors.CSRFContextView",
                "aquilify.core.messages.context_processors.Messages",
                "aquilify.template.context_processors.RequestContext"
            ],
            "extensions": [
                "aquilify.template.extensions.URLConstructor"
            ]
        }
    }
]

# CSRF Configuration...

# visit :: http://aquilify.vvfin.in/security/csrf

CSRF = [
    {
        "BACKEND": "aquilify.security.csrf.CSRF",
        "OPTIONS": {
            "secret_key": SECRET_KEY,
            "ip_verification": True,
            "trusted_ips": ["*"],
            "protected_methods": ["POST", "PUT", "DELETE"],
            "cookie_options": {"httponly": True, "secure": True, "samesite": 'strict'}
        }
    }
]

# Middleware configuration...

# You can pass either a string or a dictionary. For example:
# MIDDLEWARE = [
#    "aquilify.middlewares.XFrameOptionsMiddleware", # String middleware with default configuration.
#    {
#        "origin": "aquilify.middlewares.GzipMiddleware",
#        "order": 1,
#        "condition": "Your condition",
#        "__init__": "app",
#        "other arguments..."
#    }
# ]

# At present, we exclusively support the current app instance. This functionality dynamically adds the app instance to your class '__init__'.
# For additional details, please visit: http://aquilify.vvfin.in/middlewares/core

# The core middleware offers numerous benefits but also comes with certain exceptions. Therefore, we highly recommend using middleware hooks.
# Unlike core middleware, middleware hooks have the capability to intercept both before and after the request,
# as well as before and after the response, without any exceptions. This means you can modify the request before serving it.

MIDDLEWARES = [
    "aquilify.middlewares.XFrameOptionsMiddleware",
    "aquilify.middlewares.GzipMiddleware",
    "aquilify.middlewares.StaticMiddleware",
    "aquilify.middlewares.TrustedhostMiddleware",
    "aquilify.middlewares.CSRFMiddleware",
    "aquilify.middlewares.MediaMiddleware",
    "aquilify.middlewares.ConditionalGetMiddleware",
    "aquilify.core.messages.middleware.MessageMiddleware",
]

# stage handlers are the middleware hooks with advance features and configurations...
# we highly recommend users to use the middleware hooks in place of using regular middleware.
# For more info visit :: http://aquilify.vvfin.in/middlewares/hooks

STAGE_HANDLERS = [
    { "origin": "aquilify.core.backend.sessions.InMemorySessionBeforeStage", "stage": "before"},
    { "origin": "aquilify.core.messages.middleware.MessageBeforeStage", "stage": "before" }
]

ASGI_MIDDLEWARES = [ ]

## Database Configuration..

DATABASE = {
    "default": {
        "ENGINE": "aquilify.orm.Sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
        "MODELS": [ ]
    }
}

### Sessions Configuration....

# SESSIONS_CONFIG is a list of dictionaries containing session configuration settings for Aquilify.

# Memory backend configuration:
# - "backend": "memory" specifies the memory backend for session storage.
#   Configuration includes:
#   - "secret_key": SECRET_KEY uses the SECRET_KEY variable for session encryption.
#   - "session_lifetime": 30 sets the session lifetime in minutes for the memory backend.

# For further details on session configurations, refer to Aquilify documentation.

SESSION_BACKEND = 'memory'
SESSION_SECRET_KEY = SECRET_KEY

SESSION_MAX_AGE = 1800
SESSION_LIFETIME = 30
SESSION_SECURE = True
SESSION_HTTPONLY = True
SESSION_SAMESITE = "Lax"

SESSION_COOKIE_DOMAIN = 'localhost'
SESSION_COOKIE_HTTPONLY = True
SESSION_COOKIE_SECURE = False

SESSION_COOKIE_SAMESITE = 'Lax'


SECRET_KEY_FALLBACKS = []

# LIFESPAN Handling...

LIFESPAN_EVENTS = [ ]

# GzipMiddleware Configuration

# GZIP_COMPRESSION_LEVEL: Set the Gzip compression level to 7 for optimal compression.

# GZIP_COMPRESSION_CONTENT_TYPES: List of content types that will undergo Gzip compression.
# Supported content types include:
# - 'text/html'
# - 'text/css'
# - 'application/javascript'
# - 'application/json'
# - 'image/svg+xml'
# - 'application/xml'

# GZIP_IGNORE_CONTENT_LENGHT: Specifies whether to ignore content length for Gzip compression.
# Set to False to consider content length.

# GZIP_CONTENT_ENCODING: Indicates the content encoding to be used, in this case, set to ["gzip"].

# GZIP_EXCLUDE_PATHS: List of paths to be excluded from Gzip compression.
# Currently empty [] - add paths to exclude from compression.

# GZIP_COMPRESSION_FUNCTION: Pass the path to a custom compression function.
# Example usage: {"compression.mycustomfunc"}
# This function can be utilized for specific compression requirements.

# Ensure these settings align with the optimization and compression needs of your application.

GZIP_COMPRESSION_LEVEL = 7
GZIP_COMPRESSION_CONTENT_TYPES = ['text/html','text/css','application/javascript','application/json', 'image/svg+xml', 'application/xml']
GZIP_IGNORE_CONTENT_LENGHT = False
GZIP_CONTENT_ENCODING = ["gzip"]
GZIP_EXCLUDE_PATHS = []
GZIP_COMPRESSION_FUNCTION = {} # pass the path to cutom compression function... e.g, {"compression.mycustomfunc"}

# SIGNING BACKEND

SIGNING_BACKEND = "aquilify.core.signing.TimestampSigner"


STORAGE_BACKEND = [
    {
        "static": [
            "aquilify.core.files.storage.FileSystemStorage"
        ]
    },
    {
        "sessions": {
            "memory": "aquilify.core.backend.sessions.storage.InMemorySessionStorage",
            "cookie": "aquilify.core.backend.sessions.storage.CookieSessionStorage"
        }
    }
]

MESSAGE_STORAGE = 'aquilify.core.messages.storage.fallback.FallbackStorage'

ENVIROMENT = {
    'lxenviroment': ['packlib'] # add all the .lxe file in this list
}
"""


from http import cookies

SimpleCookie = cookies.SimpleCookie


def parse_cookie(cookie):
    cookiedict = {}
    for chunk in cookie.split(";"):
        if "=" in chunk:
            key, val = chunk.split("=", 1)
        else:
            key, val = "", chunk
        key, val = key.strip(), val.strip()
        if key or val:
            cookiedict[key] = cookies._unquote(val)
    return cookiedict



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

with open('LICENSE', 'r', encoding='utf-8') as f:
    license_text = f.read()

setup(
    name='aquilify',
    version='1.15',
    description="Aquilify is an ASGI (Asynchronous Server Gateway Interface) framework designed to facilitate the development of web applications with Python. It enables efficient handling of HTTP requests, WebSocket connections, middleware processing, and exception handling in an asynchronous environment.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Pawan kumar',
    author_email='embrakeproject@gmail.com',
    url='https://github.com/embrake/aquilify/',
    packages=find_packages(),
    keywords=['web framework', 'Python web development', 'user-friendly', 'high-level', 'ASGI', 'backend'],
    license='BSD-3-Clause',
    install_requires=['anyio'], 
    extras_require={
        'full': [
            'jinja2',
            'aiofiles',
            'python-multipart',
            'markupsafe'
        ],
        "sec": [
            "aiofiles",
            "markupsafe"
        ],
        "template": [
            "jinja2",
            "aiofiles"
        ]
    }, 
    entry_points={
        'console_scripts': [
            'aquilify = aquilify.build:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)


import importlib.util
import logging

class LoggingConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_logging_config(self):
        try:
            settings = self._load_settings_module()
            file_path = getattr(settings, 'LOG_FILE_PATH', None)
            user_agent = getattr(settings, 'LOG_USER_AGENT', True)
            client_ip = getattr(settings, 'LOG_CLIENT_IP', True)
            format = getattr(settings, 'LOG_FORMAT', None)
            log_size = getattr(settings, 'MAX_LOG_SIZE', 10 * 1024 * 1024)
            backup_count = getattr(settings, 'BACKUP_COUNT', 5)
            log_level = getattr(settings, 'LOG_LEVEL', logging.INFO)
            output_stream = getattr(settings, 'OUTPUT_STREAM', 'file')
            append_logs = getattr(settings, 'APPEND_LOGS', False)
            timestamp_format = getattr(settings, 'TIMESTAMP_FORMAT', "%Y-%m-%d %H:%M:%S")
            log_response_time = getattr(settings, 'LOG_RESPONSE_TIME', True)
            log_pattern_url = getattr(settings, 'LOG_URL_PATTERN', None)
            
            return {
                "file_path": file_path, "user_agent": user_agent, "client_ip": client_ip, "format": format, "log_size": log_size,
                "backup_count": backup_count, "log_level": log_level, "output_stream": output_stream, "append_logs": append_logs,
                "timestamp_format": timestamp_format, "log_response_time": log_response_time, "log_pattern_url": log_pattern_url
            }
        
        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings


    def fetch(self):
        data = self._fetch_logging_config()
        return data


import time
from typing import Callable, Awaitable, Dict, List, Optional, Tuple
from ..wrappers import Request, Response
from ..settings.ratelimit import RateLimitConfigSettings
from ..exception.base import TooManyRequests

_settings = RateLimitConfigSettings().fetch()

class RateLimiter:
    def __init__(
        self
    ):
        self.default_limit: int = _settings.get('limit') or 60
        self.default_window: int = _settings.get('window') or 60
        self.exempt_paths: List[str] = _settings.get('exempt_paths') or []
        self.custom_headers: Dict[str, str] = _settings.get('headers') or {}
        self.dynamic_limits: Dict[str, Callable[[int, int], Tuple[int, int]]] = _settings.get('dynamic_limits') or {}
        self.concurrency_control: bool = _settings.get('concurrency_control') or False
        self.rate_limit_expiry: Dict[str, int] = _settings.get('expiry') or {}
        self.monitoring_callback: Optional[Callable[[str, int, int, int], Awaitable[None]]] = _settings.get('monitoring_callback') or None
        self.whitelist: List[str] = _settings.get('whitelist') or []
        self.blacklist: List[str] = _settings.get('blacklist') or []
        self.requests: Dict[str, List[float]] = {}
        self.concurrent_requests: Dict[str, int] = {}
        self.rate_limit_by_endpoint: Dict[str, Tuple[int, int]] = _settings.get('endpoint') or {}
        self.rate_limit_adjustments: Dict[str, Callable[[int, int], Tuple[int, int]]] = _settings.get('adjustments') or {}
        self.rate_limit_expiry_policies: Dict[str, Callable[[int, int], int]] = _settings.get('expiry_policies') or {}
        self.ip_rate_limits: Dict[str, Tuple[int, int]] = _settings.get('ip_rate_limits') or {}
        self.rate_limit_bypass_tokens: Dict[str, List[str]] = _settings.get('bypass_token') or {}
        self.advanced_monitoring_report: Optional[Callable[[str, int, int, int], Awaitable[None]]] = _settings.get('advance_monitoring_and_reporting') or None
        self.granular_rate_limits: Dict[str, Tuple[int, int]] = _settings.get('granular') or {}
        self.automated_rate_limit_adjustment: Optional[Callable[[str, int, int], Awaitable[Tuple[int, int]]]] = _settings.get('automated_adjustments') or None
        self.granular_rate_limit_callback: Optional[Callable[[str, str], Awaitable[Tuple[int, int]]]] = _settings.get('granular_callback') or None
        self.cache_size = 1000 
        self.rate_limit_cache = {}
        self.rate_limits = {}

    async def __call__(self, request: Request, response: Response) -> Response:
        client_ip: str = request.remote_addr
        current_time: float = time.time()
        endpoint: str = request.path

        if self.is_exempt_path(endpoint) or self.is_whitelisted(client_ip) or await self.is_bypass_token(request) or client_ip in self.blacklist:
            return response

        limit, window = self.get_rate_limit(client_ip, endpoint)
        limit, window = self.apply_rate_limit_adjustments(client_ip, endpoint, limit, window)

        if self.is_concurrency_control_enabled(client_ip) and self.is_concurrency_exceeded(client_ip):
            return TooManyRequests('Concurrency limit exceeded')

        if not self.track_request(client_ip, current_time, window, limit):
            error_response = self.create_rate_limit_exceeded_response(client_ip, limit, window, current_time)
            return error_response

        if self.is_rate_limit_expired(client_ip, current_time, window):
            self.clear_rate_limit(client_ip)

        if self.monitoring_callback:
            await self.monitoring_callback(client_ip, limit, len(self.requests[client_ip]), window)

        if self.advanced_monitoring_report:
            advanced_monitoring_data = self.generate_advanced_monitoring_data(client_ip, limit, len(self.requests[client_ip]), window, response, request)
            await self.advanced_monitoring_report(advanced_monitoring_data)

        if self.granular_rate_limit_callback:
            limit, window = await self.granular_rate_limit_callback(client_ip, endpoint)
            if limit and window:
                self.rate_limits[client_ip] = (limit, window)

        return response

    def is_exempt_path(self, endpoint: str) -> bool:
        return endpoint in self.exempt_paths

    def is_whitelisted(self, client_ip: str) -> bool:
        return client_ip in self.whitelist

    async def is_bypass_token(self, request: Request) -> bool:
        token = request.headers.get("x-bypass-token")
        return any(token in tokens for tokens in self.rate_limit_bypass_tokens.values())

    def get_rate_limit(self, client_ip: str, endpoint: str) -> Tuple[int, int]:
        # Check the cache first
        cached_rate_limit = self.rate_limit_cache.get((client_ip, endpoint))
        if cached_rate_limit:
            return cached_rate_limit

        if endpoint in self.rate_limit_by_endpoint:
            rate_limit = self.rate_limit_by_endpoint[endpoint]
        elif client_ip in self.rate_limits:
            rate_limit = self.rate_limits[client_ip]
        elif client_ip in self.ip_rate_limits:
            rate_limit = self.ip_rate_limits[client_ip]
        elif endpoint in self.granular_rate_limits:
            rate_limit = self.granular_rate_limits[endpoint]
        else:
            rate_limit = (self.default_limit, self.default_window)

        if self.automated_rate_limit_adjustment:
            limit, window = self.automated_rate_limit_adjustment(client_ip, rate_limit[0], rate_limit[1])
            rate_limit = (limit, window)

        # Update the cache
        self.rate_limit_cache[(client_ip, endpoint)] = rate_limit

        return rate_limit

    def apply_rate_limit_adjustments(
        self, client_ip: str, endpoint: str, limit: int, window: int
    ) -> Tuple[int, int]:
        if endpoint in self.rate_limit_adjustments:
            return self.rate_limit_adjustments[endpoint](limit, window)
        if client_ip in self.dynamic_limits:
            return self.dynamic_limits[client_ip](limit, window)
        return limit, window

    def is_concurrency_control_enabled(self, client_ip: str) -> bool:
        return self.concurrency_control

    def is_concurrency_exceeded(self, client_ip: str) -> bool:
        if client_ip not in self.concurrent_requests:
            self.concurrent_requests[client_ip] = 0
        return self.concurrent_requests[client_ip] >= self.default_limit

    def track_request(self, client_ip: str, current_time: float, window: int, limit: int) -> bool:
        if client_ip not in self.requests:
            self.requests[client_ip] = [current_time]
        else:
            client_requests = self.requests[client_ip]
            while client_requests and client_requests[0] < current_time - window:
                client_requests.pop(0)
            if len(client_requests) < limit:
                client_requests.append(current_time)
                return True
            return False
        return True

    def create_rate_limit_exceeded_response(
        self, client_ip: str, limit: int, window: int, current_time: float
    ) -> Response:
        error_response = TooManyRequests("Limit Exceeded")
        for header, value in self.custom_headers.items():
            error_response.headers[header] = value
        error_response.headers['X-RateLimit-Limit'] = str(limit)
        error_response.headers['X-RateLimit-Remaining'] = str(max(0, limit - len(self.requests.get(client_ip, []))))
        error_response.headers['X-RateLimit-Reset'] = str(int(current_time + window))
        return error_response

    def is_rate_limit_expired(self, client_ip: str, current_time: float, window: int) -> bool:
        if client_ip in self.rate_limit_expiry and (current_time - self.rate_limit_expiry[client_ip]) >= window:
            return True
        return False

    def clear_rate_limit(self, client_ip: str):
        self.rate_limits.pop(client_ip, None)
        self.requests.pop(client_ip, None)

    def get_rate_limit_expiry_time(self, client_ip: str, endpoint: str) -> int:
        if endpoint in self.rate_limit_expiry_policies:
            return self.rate_limit_expiry_policies[endpoint](client_ip)
        return self.rate_limit_expiry.get(client_ip, 0)
    
    def generate_advanced_monitoring_data(self, client_ip: str, limit: int, request_count: int, window: int, response: Response, request: Request):
        advanced_monitoring_data = {
            'client_ip': client_ip,
            'limit': limit,
            'requests': request_count,
            'window': window,
        }

        if response:
            advanced_monitoring_data['response_status'] = response.status_code
            advanced_monitoring_data['response_headers'] = dict(response.headers)

        if request:
            advanced_monitoring_data['request_method'] = request.method
            advanced_monitoring_data['request_path'] = request.path
            advanced_monitoring_data['request_headers'] = dict(request.headers)

        return advanced_monitoring_data




import os
import ast
import importlib

class SettingsImportError(Exception):
    pass

class CompressionSetting:
    def _compression_settings():
        file_path = "./settings.py"
        try:
            with open(file_path, 'r') as file:
                parsed = ast.parse(file.read(), filename=file_path)

            variables_to_import = [
                'GZIP_COMPRESSION_LEVEL',
                'GZIP_COMPRESSION_CONTENT_TYPES',
                'GZIP_IGNORE_CONTENT_LENGHT',
                'GZIP_CONTENT_ENCODING',
                'GZIP_EXCLUDE_PATHS',
                'GZIP_COMPRESSION_FUNCTION'
            ]

            imported_settings = {}
            for node in parsed.body:
                if isinstance(node, ast.Assign) and len(node.targets) == 1:
                    target = node.targets[0]
                    if isinstance(target, ast.Name) and target.id in variables_to_import:
                        if target.id == 'GZIP_COMPRESSION_FUNCTION':
                            value = ast.literal_eval(node.value)
                            imported_functions = {}

                            for func_name, func_path in value.items():
                                try:
                                    module_name, _, func_name = func_path.rpartition('.')
                                    module = importlib.import_module(module_name)
                                    imported_functions[func_name] = getattr(module, func_name)
                                except AttributeError:
                                    raise SettingsImportError(f"Function or class '{func_name}' not found in module '{module_name}'")
                                except Exception as e:
                                    raise SettingsImportError(f"Error importing function '{func_name}' from '{func_path}': {e}")

                            imported_settings[target.id] = imported_functions
                        else:
                            value = ast.literal_eval(node.value)
                            imported_settings[target.id] = value

            return imported_settings
        except FileNotFoundError:
            raise SettingsImportError(f"File not found at path '{file_path}'")
        except SyntaxError as se:
            raise SettingsImportError(f"Syntax error in '{file_path}': {se}")
        except Exception as e:
            raise SettingsImportError(f"Error importing settings from '{file_path}': {e}")



ROOT = """# ROOT PIPELINE: @noql -> 3691
# Reference: http://aquilify.vvfin.in/pipeline/root

# Warning: Avoid altering this __root__.py file; it serves as the main tunnel for your application.
# To prevent pipeline errors, refrain from interfering with or renaming this file.

from aquilify.core import Aquilify

# Importing the core project of Aquilify.

__instance__ = Aquilify()

# Warning: If using uvicorn or other ASGI Servers, refrain from using __file__ or __instance__ to run the server.
# Warning: Changing the default __instance__ name might lead to pipeline errors. Maintain the default naming convention to prevent issues.

# Visit http://aquilify.vvfin.in/pipeline/study for comprehensive information on customizing or modifying the pipeline.

# For further assistance, contact us at control@vvfin.in or confict.con@gmail.com.

# This __root__ file is a component of the Aquilify project governed by the BSD-4 Clause LICENSE.
"""